"""
This Python module defines abstract base class for components in the Fondant data processing
framework, providing a standardized interface for extending loading and transforming components.
The loading component is the first component that loads the initial dataset and the transform
components take care of processing, filtering and extending the data.
"""
import argparse
import json
import logging
import typing as t
from abc import abstractmethod
from distutils.util import strtobool
from pathlib import Path

import dask.dataframe as dd
import pandas as pd
from fsspec import open as fs_open

from fondant.component import (
    Component,
    DaskLoadComponent,
    DaskTransformComponent,
    DaskWriteComponent,
    PandasTransformComponent,
)
from fondant.component.data_io import DaskDataLoader, DaskDataWriter
from fondant.core.component_spec import Argument, OperationSpec
from fondant.core.manifest import Manifest, Metadata

logger = logging.getLogger(__name__)


class Executor(t.Generic[Component]):
    """
    An executor executes a Component.

    Args:
        operation_spec: The operation spec of the component to be executed.
        cache: Flag indicating whether to use caching for intermediate results.
        input_manifest_path: The path to the input manifest file.
        output_manifest_path: The path to the output manifest file.
        metadata: Components metadata dict
        user_arguments: User-defined component arguments.
        input_partition_rows: The number of rows to process in each
        partition of dataframe.
        Partitions are divided based on this number (n rows per partition).
        Set to None for no row limit.
        initialise the dask client, allowing for advanced configuration.
        previous_index: The name of the index column of the previous component.
            Used to remove all previous fields if the component changes the index
    """

    def __init__(
        self,
        operation_spec: OperationSpec,
        *,
        cache: bool,
        input_manifest_path: t.Union[str, Path],
        output_manifest_path: t.Union[str, Path],
        metadata: t.Dict[str, t.Any],
        user_arguments: t.Dict[str, t.Any],
        input_partition_rows: int,
        previous_index: t.Optional[str] = None,
    ) -> None:
        self.operation_spec = operation_spec
        self.cache = cache
        self.input_manifest_path = input_manifest_path
        self.output_manifest_path = output_manifest_path
        self.metadata = Metadata.from_dict(metadata)
        self.user_arguments = user_arguments
        self.input_partition_rows = input_partition_rows
        self.previous_index = previous_index

    @classmethod
    def from_args(cls) -> "Executor":
        """Create an executor from a passed argument containing the specification as a dict."""
        parser = argparse.ArgumentParser()
        parser.add_argument("--operation_spec", type=json.loads)
        parser.add_argument("--cache", type=lambda x: bool(strtobool(x)))
        parser.add_argument("--input_partition_rows", type=int)
        args, _ = parser.parse_known_args()

        if "operation_spec" not in args:
            msg = "Error: The --operation_spec argument is required."
            raise ValueError(msg)

        operation_spec = OperationSpec.from_dict(args.operation_spec)

        return cls.from_spec(
            operation_spec,
            cache=args.cache,
            input_partition_rows=args.input_partition_rows,
        )

    @classmethod
    def from_spec(
        cls,
        operation_spec: OperationSpec,
        *,
        cache: bool,
        input_partition_rows: int,
    ) -> "Executor":
        """Create an executor from a component spec."""
        args_dict = vars(cls._add_and_parse_args(operation_spec))

        for argument in [
            "operation_spec",
            "input_partition_rows",
            "cache",
            "consumes",
            "produces",
        ]:
            args_dict.pop(argument, None)

        input_manifest_path = args_dict.pop("input_manifest_path")
        output_manifest_path = args_dict.pop("output_manifest_path")
        metadata = args_dict.pop("metadata")
        metadata = json.loads(metadata) if metadata else {}

        return cls(
            operation_spec,
            input_manifest_path=input_manifest_path,
            output_manifest_path=output_manifest_path,
            cache=cache,
            metadata=metadata,
            user_arguments=args_dict,
            input_partition_rows=input_partition_rows,
            previous_index=operation_spec.previous_index,
        )

    @classmethod
    def _add_and_parse_args(cls, operation_spec: OperationSpec):
        parser = argparse.ArgumentParser()
        component_arguments = cls._get_component_arguments(operation_spec)

        for arg in component_arguments.values():
            if arg.name in cls.optional_fondant_arguments():
                input_required = False
                default = None
            elif arg.default is not None and arg.optional is False:
                input_required = False
                default = arg.default
            elif arg.default is None and arg.optional is True:
                input_required = False
                default = None
            else:
                input_required = True
                default = None

            parser.add_argument(
                f"--{arg.name}",
                type=arg.parser,
                required=input_required,
                default=default,
                help=arg.description,
            )

        args, _ = parser.parse_known_args()
        args.__dict__ = {
            k: v if v != "None" else None for k, v in args.__dict__.items()
        }

        return args

    @staticmethod
    def optional_fondant_arguments() -> t.List[str]:
        return []

    @staticmethod
    def _get_component_arguments(
        operation_spec: OperationSpec,
    ) -> t.Dict[str, Argument]:
        """
        Get the component arguments as a dictionary representation containing both input and output
            arguments of a component
        Args:
            operation_spec: the operation spec
        Returns:
            Input and output arguments of the component.
        """
        component_arguments: t.Dict[str, Argument] = {}
        component_arguments.update(operation_spec.args)
        return component_arguments

    @abstractmethod
    def _load_or_create_manifest(self) -> Manifest:
        """Abstract method that returns the dataset manifest."""

    @abstractmethod
    def _execute_component(
        self,
        component: Component,
        *,
        manifest: Manifest,
    ) -> t.Union[None, dd.DataFrame]:
        """
        Abstract method to execute a component with the provided manifest.

        Args:
            component: Component instance to execute
            manifest: Manifest describing the input data

        Returns:
            A Dask DataFrame containing the output data
        """

    def _write_data(
        self,
        dataframe: dd.DataFrame,
        *,
        manifest: Manifest,
    ):
        """Create a data writer given a manifest and writes out the index and subsets."""
        data_writer = DaskDataWriter(
            manifest=manifest,
            operation_spec=self.operation_spec,
        )

        data_writer.write_dataframe(dataframe)

    def _get_cache_reference_content(self) -> t.Union[str, None]:
        """
        Get the content of the cache reference file. This file contains the path to the cached
        manifest or empty string if the component is cached without producing any manifest.

        Returns:
            The content of the cache reference file.
        """
        manifest_reference_path = (
            f"{self.metadata.base_path}/{self.metadata.pipeline_name}/cache/"
            f"{self.metadata.cache_key}.txt"
        )

        try:
            with fs_open(
                manifest_reference_path,
                mode="rt",
                encoding="utf-8",
            ) as file_:
                return file_.read()

        except FileNotFoundError:
            logger.info("No matching execution for component detected")
            return None

    def _is_previous_cached(self, input_manifest: Manifest) -> bool:
        """
        Checks whether the previous component's output is cached based on its run ID.

        This function compares the run ID of the input manifest
         (representing the previous component) with the run ID of the current component metadata.
        If the run IDs are different, it indicates that the previous component's output belongs to
        another pipeline run, implying that it is cached. Otherwise, if the run IDs match, it
        suggests that the previous component was not cached and had to execute to produce the
         current output.

        Args:
            input_manifest: The manifest representing the output of the previous component.

        Returns:
            True if the previous component's output is cached, False otherwise.
        """
        previous_component_id = input_manifest.component_id

        if input_manifest.run_id == self.metadata.run_id:
            logger.info(
                f"Previous component `{previous_component_id}` is not cached."
                f" Invalidating cache for current and subsequent components",
            )
            return False

        logger.info(
            f"Previous component `{previous_component_id}` run was cached. "
            f"Cached pipeline id: {input_manifest.run_id}",
        )
        return True

    def _run_execution(
        self,
        component_cls: t.Type[Component],
        input_manifest: Manifest,
    ) -> Manifest:
        logging.info("Executing component")

        component: Component = component_cls(**self.user_arguments)

        component.consumes = self.operation_spec.operation_consumes
        component.produces = self.operation_spec.operation_produces

        state = component.setup()

        output_df = self._execute_component(
            component,
            manifest=input_manifest,
        )

        output_manifest = input_manifest.evolve(
            operation_spec=self.operation_spec,
            run_id=self.metadata.run_id,
        )
        self._write_data(dataframe=output_df, manifest=output_manifest)

        component.teardown(state)

        return output_manifest

    def execute(self, component_cls: t.Type[Component]) -> None:
        """
        Execute a component.

        Args:
            component_cls: The class of the component to execute
        """
        input_manifest = self._load_or_create_manifest()
        base_path = input_manifest.base_path
        pipeline_name = input_manifest.pipeline_name

        if self.cache and self._is_previous_cached(input_manifest):
            cache_reference_content = self._get_cache_reference_content()

            if cache_reference_content is not None:
                logger.info("Skipping component execution")

                if cache_reference_content:
                    output_manifest = Manifest.from_file(cache_reference_content)

                    logger.info(
                        f"Matching execution detected for component. The last execution of the"
                        f" component originated from `{output_manifest.run_id}`.",
                    )
                else:
                    logger.info("Component is cached without producing a manifest")
                    output_manifest = None
            else:
                output_manifest = self._run_execution(component_cls, input_manifest)

        else:
            logger.info("Caching disabled for the component")
            output_manifest = self._run_execution(component_cls, input_manifest)

        if output_manifest:
            self.upload_manifest(output_manifest, save_path=self.output_manifest_path)

        self._upload_cache_reference_content(
            base_path=base_path,
            pipeline_name=pipeline_name,
        )

    def _upload_cache_reference_content(
        self,
        base_path: str,
        pipeline_name: str,
    ):
        """
        Write the cache key containing the reference to the location of the written manifest.

        This function creates a file with the format "<cache_key>.txt" at the specified
        'manifest_save_path' to store the manifest location for future retrieval of
        cached component executions.

        Args:
            base_path: The base path of the pipeline.
            pipeline_name: The name of the pipeline.
        """
        cache_reference_path = (
            f"{base_path}/{pipeline_name}/cache/{self.metadata.cache_key}.txt"
        )

        logger.info(
            f"Writing cache key with manifest reference to {cache_reference_path}",
        )

        with fs_open(
            cache_reference_path,
            mode="wt",
            encoding="utf-8",
            auto_mkdir=True,
        ) as file_:
            file_.write(self.cache_reference_content)

    @property
    def cache_reference_content(self) -> str:
        return str(self.output_manifest_path)

    def upload_manifest(self, manifest: Manifest, save_path: t.Union[str, Path]):
        """
        Uploads the manifest to the specified destination.

        Args:
            manifest: The Manifest object to be uploaded.
            save_path: The path where the Manifest object will be saved.

        """
        Path(save_path).parent.mkdir(parents=True, exist_ok=True)
        manifest.to_file(save_path)
        logger.info(f"Saving output manifest to {save_path}")


class DaskLoadExecutor(Executor[DaskLoadComponent]):
    """Base class for a Fondant load component."""

    def _is_previous_cached(self, input_manifest: Manifest) -> bool:
        return True

    @staticmethod
    def optional_fondant_arguments() -> t.List[str]:
        return ["input_manifest_path", "input_partition_rows"]

    def _load_or_create_manifest(self) -> Manifest:
        return Manifest.create(
            pipeline_name=self.metadata.pipeline_name,
            base_path=self.metadata.base_path,
            run_id=self.metadata.run_id,
            component_id=self.metadata.component_id,
            cache_key=self.metadata.cache_key,
        )

    def _execute_component(
        self,
        component: DaskLoadComponent,
        *,
        manifest: Manifest,
    ) -> dd.DataFrame:
        """This function loads the initial dataframe using the user-provided `load` method.

        Returns:
            A `dd.DataFrame` instance with initial data.
        """
        return component.load()


class TransformExecutor(Executor[Component]):
    """Base class for a Fondant transform component."""

    def _load_or_create_manifest(self) -> Manifest:
        return Manifest.from_file(self.input_manifest_path)

    def _execute_component(
        self,
        component: Component,
        *,
        manifest: Manifest,
    ) -> dd.DataFrame:
        raise NotImplementedError


class DaskTransformExecutor(TransformExecutor[DaskTransformComponent]):
    def _execute_component(
        self,
        component: DaskTransformComponent,
        *,
        manifest: Manifest,
    ) -> dd.DataFrame:
        """
        Load the data based on the manifest using a DaskDataloader and call the transform method to
        process it.

        Returns:
            A `dd.DataFrame` instance with updated data based on the applied data transformations.
        """
        data_loader = DaskDataLoader(
            manifest=manifest,
            operation_spec=self.operation_spec,
            input_partition_rows=self.input_partition_rows,
        )
        dataframe = data_loader.load_dataframe()
        return component.transform(dataframe)


class PandasTransformExecutor(TransformExecutor[PandasTransformComponent]):
    @staticmethod
    def optional_fondant_arguments() -> t.List[str]:
        return ["input_manifest_path", "input_partition_rows"]

    @staticmethod
    def wrap_transform(
        transform: t.Callable,
        *,
        operation_spec: OperationSpec,
    ) -> t.Callable:
        """Factory that creates a function to wrap the component transform function. The wrapper:
        - Skips the transformation if the received partition is empty
        - Removes extra columns from the returned dataframe which are not defined in the component
          spec `produces` section
        - Sorts the columns from the returned dataframe according to the order in the component
          spec `produces` section to match the order in the `meta` argument passed to Dask's
          `map_partitions`.

        Args:
            transform: Transform method to wrap
            operation_spec: Operation specification to base behavior on
        """

        def wrapped_transform(dataframe: pd.DataFrame) -> pd.DataFrame:
            # Columns of operation specification
            columns = [
                name for name, field in operation_spec.operation_produces.items()
            ]

            if not dataframe.empty:
                dataframe = transform(dataframe)
            else:
                logger.info("Received empty partition, skipping transformation.")

            # Drop columns not in specification
            return dataframe[columns]

        return wrapped_transform

    def _execute_component(
        self,
        component: PandasTransformComponent,
        *,
        manifest: Manifest,
    ) -> dd.DataFrame:
        """
        Load the data based on the manifest using a DaskDataloader and call the component's
        transform method for each partition of the data.

        Returns:
            A `dd.DataFrame` instance with updated data based on the applied data transformations.
        """
        data_loader = DaskDataLoader(
            manifest=manifest,
            operation_spec=self.operation_spec,
            input_partition_rows=self.input_partition_rows,
        )
        dataframe = data_loader.load_dataframe()

        # Create meta dataframe with expected format
        meta_dict = {"id": pd.Series(dtype="object")}
        for field_name, field in self.operation_spec.operation_produces.items():
            meta_dict[field_name] = pd.Series(dtype=pd.ArrowDtype(field.type.value))
        meta_df = pd.DataFrame(meta_dict).set_index("id")

        wrapped_transform = self.wrap_transform(
            component.transform,
            operation_spec=self.operation_spec,
        )

        # Call the component transform method for each partition
        dataframe = dataframe.map_partitions(
            wrapped_transform,
            meta=meta_df,
        )

        # Clear divisions if component spec indicates that the index is changed
        if self.previous_index is not None:
            dataframe.clear_divisions()

        return dataframe


class DaskWriteExecutor(Executor[DaskWriteComponent]):
    """Base class for a Fondant write component."""

    @staticmethod
    def optional_fondant_arguments() -> t.List[str]:
        return ["input_partition_rows", "output_manifest_path"]

    def _load_or_create_manifest(self) -> Manifest:
        return Manifest.from_file(self.input_manifest_path)

    def _execute_component(
        self,
        component: DaskWriteComponent,
        *,
        manifest: Manifest,
    ) -> None:
        data_loader = DaskDataLoader(
            manifest=manifest,
            operation_spec=self.operation_spec,
            input_partition_rows=self.input_partition_rows,
        )
        dataframe = data_loader.load_dataframe()
        component.write(dataframe)

    def _write_data(self, dataframe: dd.DataFrame, *, manifest: Manifest):
        """Create a data writer given a manifest and writes out the index and subsets."""

    def upload_manifest(self, manifest: Manifest, save_path: t.Union[str, Path]):
        pass

    @property
    def cache_reference_content(self) -> str:
        return ""


class ExecutorFactory:
    def __init__(self, component: t.Type[Component]):
        self.component = component
        self.component_executor_mapping: t.Dict[str, t.Type[Executor]] = {
            "DaskLoadComponent": DaskLoadExecutor,
            "DaskTransformComponent": DaskTransformExecutor,
            "DaskWriteComponent": DaskWriteExecutor,
            "PandasTransformComponent": PandasTransformExecutor,
        }

    def get_executor(self) -> Executor:
        component_type = self.component.__bases__[0].__name__
        try:
            executor = self.component_executor_mapping[component_type].from_args()
        except KeyError:
            msg = (
                f"The component `{self.component.__name__}` of type `{component_type}` has no"
                f" corresponding executor.\n "
                f"Component executor mapping:"
                f" {json.dumps(self.component_executor_mapping, indent=4)}"
            )
            raise ValueError(msg)
        return executor



"""This module defines classes to represent a Fondant Pipeline."""
import copy
import datetime
import hashlib
import inspect
import json
import logging
import re
import typing as t
from collections import OrderedDict
from dataclasses import asdict, dataclass
from pathlib import Path

try:
    from importlib.resources import files  # type: ignore
except ImportError:
    from importlib_resources import files  # type: ignore

import pyarrow as pa

from fondant.component import BaseComponent
from fondant.core.component_spec import ComponentSpec, OperationSpec
from fondant.core.exceptions import (
    InvalidLightweightComponent,
    InvalidPipelineDefinition,
)
from fondant.core.manifest import Manifest
from fondant.core.schema import Field
from fondant.pipeline import Image, LightweightComponent

logger = logging.getLogger(__name__)

VALID_ACCELERATOR_TYPES = [
    "GPU",
    "TPU",
]

# Taken from https://github.com/googleapis/python-aiplatform/blob/main/google/cloud/aiplatform_v1/types
# /accelerator_type.py
VALID_VERTEX_ACCELERATOR_TYPES = [
    "ACCELERATOR_TYPE_UNSPECIFIED",
    "NVIDIA_TESLA_K80",
    "NVIDIA_TESLA_P100",
    "NVIDIA_TESLA_V100",
    "NVIDIA_TESLA_P4",
    "NVIDIA_TESLA_T4",
    "NVIDIA_TESLA_A100",
    "NVIDIA_A100_80GB",
    "NVIDIA_L4",
    "TPU_V2",
    "TPU_V3",
    "TPU_V4_POD",
]


@dataclass
class Resources:
    accelerator_name: t.Optional[str] = None
    accelerator_number: t.Optional[int] = None
    cpu_request: t.Optional[str] = None
    cpu_limit: t.Optional[str] = None
    memory_request: t.Optional[str] = None
    memory_limit: t.Optional[str] = None
    node_pool_label: t.Optional[str] = None
    node_pool_name: t.Optional[str] = None
    instance_type: t.Optional[str] = None

    """
    Class representing the resources to assign to a Fondant Component operation in a Fondant
    Pipeline.

       Arguments:
           number_of_accelerators: The number of accelerators to assign to the operation (GPU, TPU)
           accelerator_name: The name of the accelerator to assign. If you're using a cluster setup
             on GKE, select "GPU" for GPU or "TPU" for TPU. Make sure
             that you select a nodepool with the available hardware. If you're running the
             pipeline on Vertex, then select one of the machines specified in the list of
             accelerators here https://cloud.google.com/vertex-ai/docs/reference/rest/v1/MachineSpec.
           node_pool_label: The label of the node pool to which the operation will be assigned.
           node_pool_name: The name of the node pool to which the operation will be assigned.
           cache: Set to False to disable caching, True by default.
           cpu_request: the memory requested by the component. The value
            should be a string which can be a number or a number followed by “m”, which means
            1/1000.
           cpu_limit: the maximum amount of CPU that can be used by the component. The value
            should be a string which can be a number or a number followed by “m”, which means
            1/1000.
           memory_request: the memory requested by the component. The value  can be a number or a
             number followed by one of “E”, “P”, “T”, “G”, “M”, “K”.
           memory_limit: the maximum memory that can be used by the component. The value  can be a
            number or a number followed by one of “E”, “P”, “T”, “G”, “M”, “K”.
           instancy_type: the instancy type of the component.
       """

    def __post_init__(self):
        """Validate the resources."""
        if bool(self.node_pool_label) != bool(self.node_pool_name):
            msg = "Both node_pool_label and node_pool_name must be specified or both must be None."
            raise InvalidPipelineDefinition(
                msg,
            )

        if bool(self.accelerator_number) != bool(self.accelerator_name):
            msg = (
                "Both number of accelerators and accelerator name must be specified or both must"
                " be None."
            )
            raise InvalidPipelineDefinition(
                msg,
            )

    def to_dict(self) -> t.Dict[str, t.Any]:
        """Return a dictionary representation of the Resources object."""
        return asdict(self)


class ComponentOp:
    """
    Class representing an operation for a Fondant Component in a Fondant Pipeline. An operation
    is a representation of a function that will be executed as part of a pipeline.

    Arguments:
        name_or_path: The name of a reusable component, or the path to the directory containing a
            custom component.
        arguments: A dictionary containing the argument name and value for the operation.
        input_partition_rows: The number of rows to load per partition. Set to override the
        automatic partitioning
        resources: The resources to assign to the operation.

    Note:
        - A Fondant Component operation is created by defining a Fondant Component and its input
          arguments.
    """

    COMPONENT_SPEC_NAME = "fondant_component.yaml"

    def __init__(
        self,
        image: Image,
        component_spec: ComponentSpec,
        *,
        consumes: t.Optional[t.Dict[str, str]] = None,
        produces: t.Optional[t.Dict[str, t.Union[str, pa.DataType]]] = None,
        arguments: t.Optional[t.Dict[str, t.Any]] = None,
        input_partition_rows: t.Optional[t.Union[str, int]] = None,
        cache: t.Optional[bool] = True,
        resources: t.Optional[Resources] = None,
        component_dir: t.Optional[Path] = None,
        dataset_fields: t.Optional[t.Mapping[str, Field]] = None,
    ) -> None:
        self.image = image
        self.component_spec = component_spec
        self.input_partition_rows = input_partition_rows
        self.cache = self._configure_caching_from_image_tag(cache)
        self.component_dir = component_dir

        if consumes is None:
            consumes = self._infer_consumes(component_spec, dataset_fields)
        consumes = self._validate_consumes(consumes, component_spec, dataset_fields)

        self.operation_spec = OperationSpec(
            self.component_spec,
            consumes=consumes,
            produces=produces,
        )

        self.arguments = arguments or {}
        self.arguments.update(
            {
                key: value
                for key, value in {
                    "input_partition_rows": input_partition_rows,
                    "cache": self.cache,
                    "operation_spec": self.operation_spec.to_json(),
                }.items()
                if value is not None
            },
        )

        self.arguments.setdefault("operation_spec", self.operation_spec.to_json())

        self.resources = resources or Resources()

    @classmethod
    def from_component_yaml(cls, path, fields=None, **kwargs) -> "ComponentOp":
        if cls._is_custom_component(path):
            component_dir = Path(path)
        else:
            component_dir = cls._get_registry_path(str(path))

        component_spec = ComponentSpec.from_file(
            component_dir / cls.COMPONENT_SPEC_NAME,
        )

        image = Image(
            base_image=component_spec.image,
        )

        return cls(
            image=image,
            component_spec=component_spec,
            component_dir=component_dir,
            dataset_fields=fields,
            **kwargs,
        )

    @classmethod
    def _infer_consumes(
        cls,
        component_spec,
        dataset_fields,
    ) -> t.Union[t.Optional[t.Dict[str, str]], t.Optional[t.Dict[str, pa.DataType]]]:
        """Infer the consumes section of the component spec."""
        if component_spec.consumes_is_defined is False:
            msg = (
                "The consumes section of the component spec is not defined. "
                "Can not infer consumes of the OperationSpec. Please define a consumes section "
                "in the dataset interface. "
            )
            logger.info(msg)
            return None

        # Component has consumes and additionalProperties, we will load all dataset columns
        if (
            component_spec.consumes_is_defined
            and component_spec.consumes_additional_properties
        ):
            if dataset_fields is None:
                logger.info(
                    "The dataset fields are not defined. Cannot infer consumes.",
                )
                return None

            return {k: v.type.value for k, v in dataset_fields.items()}

        # Component has consumes and no additionalProperties, we will load only the columns defined
        # in the component spec
        return {k: v.type.value for k, v in component_spec.consumes.items()}

    @classmethod
    def _validate_consumes(
        cls,
        consumes: t.Optional[t.Dict[str, str]],
        component_spec: ComponentSpec,
        dataset_fields: t.Optional[t.Mapping[str, Field]],
    ) -> t.Union[t.Optional[t.Dict[str, str]], t.Optional[t.Dict[str, pa.DataType]]]:
        """
        Validate the consumes of the component spec.
        Every column in the consumes should be present in the dataset fields and in the
        ComponentSpec. Except if additionalProperties is set to True in the ComponentSpec.
        In that case, we will infer the type from the dataset fields.
        """
        if consumes is None or dataset_fields is None:
            return consumes

        validated_consumes = copy.deepcopy(consumes)

        for operations_column_name, dataset_column_name_or_type in consumes.items():
            # Dataset column name is part of the dataset fields
            if (
                isinstance(dataset_column_name_or_type, str)
                and dataset_column_name_or_type not in dataset_fields.keys()
            ):
                msg = (
                    f"The dataset does not contain the column {dataset_column_name_or_type} "
                    f"required by the component {component_spec.name}."
                )
                raise InvalidPipelineDefinition(msg)

            # If operations column name is not in the component spec, but additional properties
            # are true we will infer the correct type from the dataset fields
            if (
                isinstance(dataset_column_name_or_type, str)
                and operations_column_name not in component_spec.consumes.keys()
            ):
                if component_spec.consumes_additional_properties:
                    validated_consumes[operations_column_name] = dataset_fields[
                        operations_column_name
                    ].type.value
                else:
                    msg = (
                        f"Received a string value for key `{operations_column_name}` in the "
                        f"`consumes` argument passed to the operation, "
                        f"but `{operations_column_name}` is not defined in the `consumes` "
                        f"section of the component spec."
                    )
                    raise InvalidPipelineDefinition(msg)

        return validated_consumes

    @classmethod
    def from_ref(
        cls,
        ref: t.Any,
        fields: t.Optional[t.Mapping[str, Field]] = None,
        **kwargs,
    ) -> "ComponentOp":
        """Create a ComponentOp from a reference. The reference can
        be a reusable component name, a path to a custom component,
        or a python component class.

        Args:
            ref: The name of a reusable component, or the path to the directory containing
                a custom component, or a python component class.
            fields: The fields of the dataset available to the component.
            **kwargs: The provided user arguments are passed in as keyword arguments
        """
        if inspect.isclass(ref) and issubclass(ref, BaseComponent):
            if issubclass(ref, LightweightComponent):
                component_spec = ref.get_component_spec()

                operation = cls(
                    ref.image(),
                    component_spec,
                    dataset_fields=fields,
                    **kwargs,
                )
            else:
                msg = """Reference is not a valid lightweight component.
                       Make sure the component is decorated properly."""
                raise InvalidLightweightComponent(msg)

        elif isinstance(ref, (str, Path)):
            operation = cls.from_component_yaml(
                ref,
                fields,
                **kwargs,
            )
        else:
            msg = f"""Invalid reference type: {type(ref)}.
                Expected a string, Path, or a lightweight component class."""
            raise ValueError(msg)
        return operation

    def _configure_caching_from_image_tag(
        self,
        cache: t.Optional[bool],
    ) -> t.Optional[bool]:
        """
        Adjusts the caching setting based on the image tag of the component.

        If the `cache` parameter is set to `True`, this function checks the image tag of
        the component and disables caching (`cache` set to `False`) if the image tag is "latest".
        This is because using "latest" image tags can lead to unpredictable behavior due to
         image updates.

        Args:
            cache: The current caching setting. Set to `True` to enable caching, `False` to disable.

        Returns:
            The adjusted caching setting based on the image tag.

        """
        if cache is True:
            image_tag = self.component_spec.image.rsplit(":")[-1]

            if image_tag == "latest":
                logger.warning(
                    f"Component `{self.component_spec.name}` has an image tag set to latest. "
                    f"Caching for the component will be disabled to prevent"
                    f" unpredictable behavior due to images updates",
                )
                return False

        return cache

    @property
    def dockerfile_path(self) -> t.Optional[Path]:
        if not self.component_dir:
            return None
        docker_path = self.component_dir / "Dockerfile"
        return docker_path if docker_path.exists() else None

    @staticmethod
    def _is_custom_component(path_or_name: t.Union[str, Path]) -> bool:
        """Checks if name is a local path and a custom component."""
        component_dir: Path = Path(path_or_name)
        return component_dir.exists() and component_dir.is_dir()

    @staticmethod
    def _get_registry_path(name: str) -> Path:
        """Checks if name is a local path and a custom component."""
        component_dir: Path = t.cast(Path, files("fondant") / f"components/{name}")
        if not (component_dir.exists() and component_dir.is_dir()):
            msg = f"No reusable component with name {name} found."
            raise ValueError(msg)
        return component_dir

    @property
    def component_name(self) -> str:
        return self.component_spec.safe_name

    def get_component_cache_key(
        self,
        previous_component_cache: t.Optional[str] = None,
    ) -> str:
        """Calculate a cache key representing the unique identity of this ComponentOp.

        The cache key is computed based on the component specification, image hash, arguments, and
        other attributes of the ComponentOp. It is used to uniquely identify a specific instance
        of the ComponentOp and is used for caching.

        Returns:
            A cache key representing the unique identity of this ComponentOp.
        """

        def get_nested_dict_hash(input_dict):
            """Calculate the hash of a nested dictionary.

            Args:
                input_dict: The nested dictionary to calculate the hash for.

            Returns:
                The hash value (MD5 digest) of the nested dictionary.
            """
            sorted_json_string = json.dumps(input_dict, sort_keys=True)
            hash_object = hashlib.md5(sorted_json_string.encode())  # nosec
            return hash_object.hexdigest()

        operation_spec_dict = self.operation_spec.to_dict()
        image_dict = self.image.to_dict()

        arguments = (
            get_nested_dict_hash(self.arguments) if self.arguments is not None else None
        )

        component_op_uid_dict = {
            "operation_spec_hash": get_nested_dict_hash(operation_spec_dict),
            "image": get_nested_dict_hash(image_dict),
            "arguments": arguments,
            "input_partition_rows": self.input_partition_rows,
            "number_of_accelerators": self.resources.accelerator_number,
            "accelerator_name": self.resources.accelerator_name,
            "node_pool_name": self.resources.node_pool_name,
        }

        if previous_component_cache is not None:
            component_op_uid_dict["previous_component_cache"] = previous_component_cache

        return get_nested_dict_hash(component_op_uid_dict)


class Pipeline:
    """Class representing a Fondant Pipeline."""

    def __init__(
        self,
        name: str,
        *,
        base_path: str,
        description: t.Optional[str] = None,
    ):
        """
        Args:
            name: The name of the pipeline.
            base_path: The base path for the pipeline to use to store artifacts and data. This
                can be a local path or a remote path on one of the supported cloud storage
                services. The path should already exist.
            description: Optional description of the pipeline.
        """
        self.base_path = base_path
        self.name = self._validate_pipeline_name(name)
        self.description = description
        self.package_path = f"{name}.tgz"
        self._graph: t.OrderedDict[str, t.Any] = OrderedDict()
        self.task_without_dependencies_added = False

    def register_operation(
        self,
        operation: ComponentOp,
        *,
        input_dataset: t.Optional["Dataset"],
        output_dataset: t.Optional["Dataset"],
    ) -> None:
        dependencies = []
        for component_name, info in self._graph.items():
            if info["output_dataset"] == input_dataset:
                dependencies.append(component_name)

        self._graph[operation.component_name] = {
            "operation": operation,
            "dependencies": dependencies,
            "output_dataset": output_dataset,
        }

    def read(
        self,
        ref: t.Any,
        *,
        produces: t.Optional[t.Dict[str, t.Union[str, pa.DataType]]] = None,
        arguments: t.Optional[t.Dict[str, t.Any]] = None,
        input_partition_rows: t.Optional[t.Union[int, str]] = None,
        resources: t.Optional[Resources] = None,
        cache: t.Optional[bool] = True,
    ) -> "Dataset":
        """
        Read data using the provided component.

        Args:
            ref: The name of a reusable component, or the path to the directory containing
                a containerized component, or a lightweight component class.
            produces: A mapping to update the fields produced by the operation as defined in the
                component spec. The keys are the names of the fields to be received by the
                component, while the values are the type of the field, or the name of the field to
                map from the dataset.
            arguments: A dictionary containing the argument name and value for the operation.
            input_partition_rows: The number of rows to load per partition. Set to override the
            automatic partitioning
            resources: The resources to assign to the operation.
            cache: Set to False to disable caching, True by default.

        Returns:
            An intermediate dataset.
        """
        if self._graph:
            msg = "For now, at most one read component can be applied per pipeline."
            raise InvalidPipelineDefinition(
                msg,
            )

        operation = ComponentOp.from_ref(
            ref,
            produces=produces,
            arguments=arguments,
            input_partition_rows=input_partition_rows,
            resources=resources,
            cache=cache,
        )
        manifest = Manifest.create(
            pipeline_name=self.name,
            base_path=self.base_path,
            run_id=self.get_run_id(),
            component_id=operation.component_name,
        )
        dataset = Dataset(manifest, pipeline=self)

        return dataset._apply(operation)

    def sort_graph(self):
        """Sort the graph topologically based on task dependencies."""
        logger.info("Sorting pipeline component graph topologically.")
        sorted_graph = []
        visited = set()

        def depth_first_traversal(node: str):
            """
            Perform a depth-first traversal of the graph and its dependencies.

            Args:
                node: The name of the starting node for traversal.
            """
            if node not in visited:
                visited.add(node)
                for dependency in self._graph[node]["dependencies"]:
                    depth_first_traversal(dependency)
                sorted_graph.append(node)

        for graph_node in self._graph:
            depth_first_traversal(graph_node)

        self._graph = OrderedDict((node, self._graph[node]) for node in sorted_graph)

    @staticmethod
    def _validate_pipeline_name(pipeline_name: str) -> str:
        pattern = r"^[a-z0-9][a-z0-9_-]*$"
        if not re.match(pattern, pipeline_name):
            msg = f"The pipeline name violates the pattern {pattern}"
            raise InvalidPipelineDefinition(msg)
        return pipeline_name

    def get_run_id(self) -> str:
        """Get a unique run ID for the pipeline."""
        timestamp = datetime.datetime.now().strftime("%Y%m%d%H%M%S")
        return f"{self.name}-{timestamp}"

    def validate(self, run_id: str):
        """Sort and run validation on the pipeline definition.

        Args:
            run_id: run identifier

        """
        self.sort_graph()
        self._validate_pipeline_definition(run_id)

    def _validate_pipeline_definition(self, run_id: str):
        """
        Validates the pipeline definition by ensuring that the consumed and produced subsets and
        their associated fields match and are invoked in the correct order.

        Raises:
            InvalidPipelineDefinition: If a component is trying to invoke a subset that is not
             defined or created in previous components, or if an invoked subset's schema does not
              match the previously created subset definition.
            base_path: the base path where to store the pipelines artifacts
            run_id: the run id of the component
        """
        if len(self._graph.keys()) == 0:
            logger.info("No components defined in the pipeline. Nothing to validate.")
            return

        # TODO: change later if we decide to run 2 fondant pipelines after each other
        load_component = True
        load_component_name = list(self._graph.keys())[0]

        # Create initial manifest
        manifest = Manifest.create(
            pipeline_name=self.name,
            base_path=self.base_path,
            run_id=run_id,
            component_id=load_component_name,
            cache_key="42",
        )
        for operation_specs in self._graph.values():
            component_op = operation_specs["operation"]
            operation_spec = component_op.operation_spec

            if not load_component:
                # Check subset exists
                for (
                    component_field_name,
                    component_field,
                ) in operation_spec.consumes_from_dataset.items():
                    if component_field_name not in manifest.fields:
                        msg = (
                            f"Component '{component_op.component_name}' is trying to invoke the"
                            f"field '{component_field_name}', which has not been defined or created"
                            f"in the previous components. \n"
                            f"Available field names: {list(manifest.fields.keys())}"
                        )
                        raise InvalidPipelineDefinition(
                            msg,
                        )

                    # Get the corresponding manifest fields
                    manifest_field = manifest.fields[component_field_name]

                    # Check if the invoked field schema matches the current schema
                    if component_field.type != manifest_field.type:
                        msg = (
                            f"The invoked field '{component_field_name}' of the "
                            f"'{component_op.component_name}' component does not match the "
                            f"previously created field type.\n The '{manifest_field.name}' "
                            f"field is currently defined with the following type:\n"
                            f"{manifest_field.type}\nThe current component to "
                            f"trying to invoke it with this type:\n{component_field.type}"
                        )
                        raise InvalidPipelineDefinition(
                            msg,
                        )

            manifest = manifest.evolve(operation_spec, run_id=run_id)
            load_component = False

        logger.info("All pipeline component specifications match.")

    def __repr__(self) -> str:
        """Return a string representation of the FondantPipeline object."""
        return f"{self.__class__.__name__}({self._graph!r}"


class Dataset:
    def __init__(self, manifest, *, pipeline: Pipeline) -> None:
        """A class representing an intermediate dataset.

        Args:
            manifest: Manifest representing the dataset
            pipeline: The pipeline this dataset is a part of.
        """
        self.manifest = manifest
        self.pipeline = pipeline

    @property
    def fields(self) -> t.Mapping[str, Field]:
        """The fields of the manifest as an immutable mapping."""
        return dict(self.manifest.fields)

    def _apply(self, operation: ComponentOp) -> "Dataset":
        """
        Apply the provided operation to the dataset.

        Args:
            operation: The operation to apply.
        """
        evolved_manifest = self.manifest.evolve(
            operation.operation_spec,
            run_id=self.pipeline.get_run_id(),
        )
        evolved_dataset = Dataset(evolved_manifest, pipeline=self.pipeline)

        if self.pipeline is not None:
            self.pipeline.register_operation(
                operation,
                input_dataset=self,
                output_dataset=evolved_dataset,
            )

        return evolved_dataset

    def apply(
        self,
        ref: t.Any,
        *,
        consumes: t.Optional[t.Dict[str, t.Union[str, pa.DataType]]] = None,
        produces: t.Optional[t.Dict[str, t.Union[str, pa.DataType]]] = None,
        arguments: t.Optional[t.Dict[str, t.Any]] = None,
        input_partition_rows: t.Optional[t.Union[int, str]] = None,
        resources: t.Optional[Resources] = None,
        cache: t.Optional[bool] = True,
    ) -> "Dataset":
        """
        Apply the provided component on the dataset.

        Args:
            ref: The name of a reusable component, or the path to the directory containing
                a custom component, or a lightweight component class.
            consumes: A mapping to update the fields consumed by the operation as defined in the
                component spec. The keys are the names of the fields to be received by the
                component, while the values are the type of the field, or the name of the field to
                map from the input dataset.

                Suppose we have a component spec that expects the following fields:

                ```
                ...
                consumes:
                    text:
                        type: string
                    image:
                        type: binary
                ...
                ```

                To override the default mapping and specify that the 'text' field should be sourced
                from the 'custom_text' field in the input dataset, the 'consumes' mapping can be
                defined as follows:

                ```
                consumes = {
                    "text": "custom_text"
                }
                ```

                In this example, the 'text' field will be sourced from 'custom_text' and 'image'
                will be sourced from the 'image' field by default, since it's not specified in the
                custom mapping.

            produces: A mapping to update the fields produced by the operation as defined in the
                component spec. The keys are the names of the fields to be produced by the
                component, while the values are the type of the field, or the name that should be
                used to write the field to the output dataset.

                Suppose we have a component spec that expects the following fields:

                ```
                ...
                produces:
                    text:
                        type: string
                    width:
                        type: int
                ```

                To customize the field names and types during the production step, the 'produces'
                mapping can be defined as follows:

                ```
                produces = {
                    "width": "custom_width",
                }
                ```

                In this example, the 'text' field will retain as text since it is not specified in
                the custom mapping. The 'width' field will be stored with the name 'custom_width'
                in the output dataset.

                Alternatively, the produces defines the data type of the output data.

                ```
                produces = {
                    "width": pa.float32(),
                }
                ```

                In this example, the 'text' field will retain its type 'string' without specifying a
                different source, while the 'width' field will be produced as type `float` in the
                output dataset.

            arguments: A dictionary containing the argument name and value for the operation.
            input_partition_rows: The number of rows to load per partition. Set to override the
            automatic partitioning
            resources: The resources to assign to the operation.
            cache: Set to False to disable caching, True by default.

        Returns:
            An intermediate dataset.
        """
        operation = ComponentOp.from_ref(
            ref,
            fields=self.fields,
            produces=produces,
            consumes=consumes,
            arguments=arguments,
            input_partition_rows=input_partition_rows,
            resources=resources,
            cache=cache,
        )

        return self._apply(operation)

    def write(
        self,
        ref: t.Any,
        *,
        consumes: t.Optional[t.Dict[str, t.Union[str, pa.DataType]]] = None,
        arguments: t.Optional[t.Dict[str, t.Any]] = None,
        input_partition_rows: t.Optional[t.Union[int, str]] = None,
        resources: t.Optional[Resources] = None,
        cache: t.Optional[bool] = True,
    ) -> None:
        """
        Write the dataset using the provided component.

        Args:
            ref: The name of a reusable component, or the path to the directory containing
                a custom component, or a lightweight component class.
            consumes: A mapping to update the fields consumed by the operation as defined in the
                component spec. The keys are the names of the fields to be received by the
                component, while the values are the type of the field, or the name of the field to
                map from the input dataset.
            arguments: A dictionary containing the argument name and value for the operation.
            input_partition_rows: The number of rows to load per partition. Set to override the
            automatic partitioning
            resources: The resources to assign to the operation.
            cache: Set to False to disable caching, True by default.

        Returns:
            An intermediate dataset.
        """
        operation = ComponentOp.from_ref(
            ref,
            fields=self.fields,
            consumes=consumes,
            arguments=arguments,
            input_partition_rows=input_partition_rows,
            resources=resources,
            cache=cache,
        )
        self._apply(operation)



from .lightweight_component import (  # noqa
    Image,
    LightweightComponent,
    lightweight_component,
)
from .pipeline import (  # noqa
    VALID_ACCELERATOR_TYPES,
    VALID_VERTEX_ACCELERATOR_TYPES,
    ComponentOp,
    Dataset,
    Pipeline,
    Resources,
)



import inspect
import itertools
import logging
import sys
import textwrap
import typing as t
from dataclasses import asdict, dataclass
from functools import wraps
from importlib import metadata

import pyarrow as pa

from fondant.component import BaseComponent, Component
from fondant.core.component_spec import ComponentSpec
from fondant.core.schema import Type
from fondant.pipeline.argument_inference import infer_arguments

logger = logging.getLogger(__name__)

MIN_PYTHON_VERSION = (3, 9)
MAX_PYTHON_VERSION = (3, 11)


@dataclass
class Image:
    base_image: str
    extra_requires: t.Optional[t.List[str]] = None
    script: t.Optional[str] = None

    def __post_init__(self):
        if self.base_image is None:
            self.base_image = self.resolve_fndnt_base_image()

        # log info when custom image without Fondant is defined
        elif self.extra_requires and not any(
            dependency.startswith("fondant") for dependency in self.extra_requires
        ):
            msg = (
                "You are not using a Fondant default base image, and Fondant is not part of"
                "your extra requirements. Please make sure that you have installed fondant "
                "inside your container. Alternatively, you can should add Fondant to "
                "the extra requirements. \n"
                "E.g. \n"
                '@lightweight_component(..., extra_requires=["fondant"])'
            )

            logger.warning(msg)

    @staticmethod
    def resolve_fndnt_base_image():
        """Resolve the correct fndnt base image using python version and fondant version."""
        # Set python version to latest supported version
        python_version = sys.version_info
        if MIN_PYTHON_VERSION <= python_version < MAX_PYTHON_VERSION:
            python_version = f"{python_version.major}.{python_version.minor}"
        else:
            python_version = f"{MAX_PYTHON_VERSION[0]}.{MAX_PYTHON_VERSION[1]}"

        fondant_version = metadata.version("fondant")
        if fondant_version == "0.1.dev0":
            fondant_version = "dev"
        basename = "fndnt/fondant"
        return f"{basename}:{fondant_version}-py{python_version}"

    def to_dict(self):
        return asdict(self)


class LightweightComponent(BaseComponent):
    @classmethod
    def image(cls) -> Image:
        raise NotImplementedError

    @classmethod
    def consumes(cls) -> t.Optional[t.Dict[str, t.Any]]:
        pass

    @classmethod
    def produces(cls) -> t.Optional[t.Dict[str, t.Any]]:
        pass

    @classmethod
    def _get_spec_consumes(cls) -> t.Mapping[str, t.Union[str, pa.DataType, bool]]:
        """Get the consumes spec for the component."""
        consumes = cls.consumes()
        if consumes is None:
            return {"additionalProperties": True}

        return {
            k: (Type(v).to_dict() if k != "additionalProperties" else v)
            for k, v in consumes.items()
        }

    @classmethod
    def _get_spec_produces(cls) -> t.Mapping[str, t.Union[str, pa.DataType, bool]]:
        """Get the produces spec for the component."""
        produces = cls.produces()

        if produces is None:
            return {"additionalProperties": True}

        return {
            k: (Type(v).to_dict() if k != "additionalProperties" else v)
            for k, v in produces.items()
        }

    @classmethod
    def get_component_spec(cls) -> ComponentSpec:
        """Return the component spec for the component."""
        return ComponentSpec(
            name=cls.__name__,
            image=cls.image().base_image,
            description=cls.__doc__ or "lightweight component",
            consumes=cls._get_spec_consumes(),
            produces=cls._get_spec_produces(),
            args={name: arg.to_spec() for name, arg in infer_arguments(cls).items()},
        )


def lightweight_component(
    *args,
    extra_requires: t.Optional[t.List[str]] = None,
    base_image: t.Optional[str] = None,
    consumes: t.Optional[t.Dict[str, t.Any]] = None,
    produces: t.Optional[t.Dict[str, t.Any]] = None,
):
    """Decorator to enable a lightweight component."""

    def wrapper(cls):
        script = build_python_script(cls)
        image = Image(
            base_image=base_image,
            extra_requires=extra_requires,
            script=script,
        )

        def get_base_cls(cls):
            """
            Returns the BaseComponent. If the implementation inherits from several classes,
            the Fondant base class is selected. If more than one Fondant base component
            is implemented, an exception is raised.
            """
            base_component_module = inspect.getmodule(Component).__name__
            base_component_cls_list = [
                base
                for base in cls.__bases__
                if base.__module__ == base_component_module
            ]
            if len(base_component_cls_list) > 1:
                msg = (
                    f"Multiple base classes detected. Only one component should be inherited or"
                    f" implemented."
                    f"Found classes: {', '.join([cls.__name__ for cls in base_component_cls_list])}"
                )
                raise ValueError(
                    msg,
                )
            return base_component_cls_list[0]

        def validate_signatures(base_component_cls, cls_implementation):
            """
            Compare the signature of overridden methods in a class with their counterparts
            in the BaseComponent classes.
            """
            for function_name in dir(cls_implementation):
                if not function_name.startswith("__") and function_name in dir(
                    base_component_cls,
                ):
                    type_cls_implementation = inspect.signature(
                        getattr(cls_implementation, function_name, None),
                    )
                    type_base_cls = inspect.signature(
                        getattr(base_component_cls, function_name, None),
                    )
                    if type_cls_implementation != type_base_cls:
                        msg = (
                            f"Invalid function definition of function {function_name}. "
                            f"The expected function signature is {type_base_cls}"
                        )
                        raise ValueError(
                            msg,
                        )

        def validate_abstract_methods_are_implemented(cls):
            """
            Function to validate that a class has overridden every required function marked as
            abstract.
            """
            abstract_methods = [
                name
                for name, value in inspect.getmembers(cls)
                if getattr(value, "__isabstractmethod__", False)
            ]
            if len(abstract_methods) >= 1:
                msg = (
                    f"Every required function must be overridden in the LightweightComponent. "
                    f"Missing implementations for the following functions: {abstract_methods}"
                )
                raise ValueError(
                    msg,
                )

        validate_abstract_methods_are_implemented(cls)
        base_component_cls = get_base_cls(cls)
        validate_signatures(base_component_cls, cls)

        # updated=() is needed to prevent an attempt to update the class's __dict__
        @wraps(cls, updated=())
        class LightweightComponentOp(cls, LightweightComponent):
            @classmethod
            def image(cls) -> Image:
                return image

            @classmethod
            def produces(cls) -> t.Optional[t.Dict[str, pa.DataType]]:
                return produces

            @classmethod
            def consumes(cls) -> t.Optional[t.Dict[str, t.Dict[t.Any, t.Any]]]:
                return consumes

        return LightweightComponentOp

    # Call wrapper with function (`args[0]`) when no additional arguments were passed
    if args:
        return wrapper(args[0])

    return wrapper


def new_getfile(_object, _old_getfile=inspect.getfile):
    if not inspect.isclass(_object):
        return _old_getfile(_object)

    # Lookup by parent module (as in current inspect)
    if hasattr(_object, "__module__"):
        object_ = sys.modules.get(_object.__module__)
        if hasattr(object_, "__file__"):
            return object_.__file__

    # If parent module is __main__, lookup by methods (NEW)
    for name, member in inspect.getmembers(_object):
        if (
            inspect.isfunction(member)
            and _object.__qualname__ + "." + member.__name__ == member.__qualname__
        ):
            return inspect.getfile(member)

    msg = f"Source for {_object!r} not found"
    raise TypeError(msg)


def is_running_interactively():
    """Check if the code is running in an interactive environment."""
    try:
        from IPython import get_ipython

        return get_ipython() is not None
    except ModuleNotFoundError:
        return False


def build_python_script(component_cls: t.Type[Component]) -> str:
    """Build a self-contained python script for the provided component class, which will act as
    the `src/main.py` script to execute the component.
    """
    imports_source = textwrap.dedent(
        """\
        from typing import *
        import typing as t

        import dask.dataframe as dd
        import fondant
        import pandas as pd
        from fondant.component import *
        from fondant.core import *
    """,
    )

    if is_running_interactively():
        from IPython.core.magics.code import extract_symbols

        component_source = "".join(
            inspect.linecache.getlines(  # type: ignore[attr-defined]
                new_getfile(component_cls),
            ),
        )
        component_source = extract_symbols(
            component_source,
            component_cls.__name__,
        )
        component_source = component_source[0][0]

    else:
        component_source = inspect.getsource(component_cls)

    component_source = textwrap.dedent(component_source)
    component_source_lines = component_source.split("\n")

    # Removing possible decorators (can be multiline) until the class
    # definition is found
    component_source_lines = list(
        itertools.dropwhile(
            lambda x: not x.startswith("class"),
            component_source_lines,
        ),
    )

    if not component_source_lines:
        msg = (
            f'Failed to dedent and clean up the source of function "{component_cls.__name__}". '
            f"Its probably not properly indented."
        )
        raise ValueError(
            msg,
        )

    component_source = "\n".join(component_source_lines)

    return "\n\n".join([imports_source, component_source])



# fmt: off
from .component import (BaseComponent, Component, DaskLoadComponent,  # noqa
                        DaskTransformComponent, DaskWriteComponent,
                        PandasTransformComponent)

# fmt: on



import logging
import os
import shlex
import subprocess  # nosec
import typing as t
from dataclasses import asdict
from importlib.metadata import version
from pathlib import Path

import fsspec.core
import yaml
from fsspec.implementations.local import LocalFileSystem

from fondant.core.schema import CloudCredentialsMount, DockerVolume

CONTAINER = "fndnt/data_explorer"
PORT = 8501
OUTPUT_PATH = ".fondant/explorer-compose.yaml"


def _get_service_name(container: str) -> str:
    return os.path.basename(container)


def _generate_explorer_spec(
    *,
    base_path: str,
    port: int = PORT,
    container: str = CONTAINER,
    tag: t.Optional[str] = None,
    extra_volumes: t.Union[t.Optional[list], t.Optional[str]] = None,
    auth_provider: t.Optional[CloudCredentialsMount] = None,
) -> t.Dict[str, t.Any]:
    """Generate a Docker Compose specification for the Explorer App."""
    if tag is None:
        tag = version("fondant") if version("fondant") != "0.1.dev0" else "latest"

    if extra_volumes is None:
        extra_volumes = []

    if isinstance(extra_volumes, str):
        extra_volumes = [extra_volumes]

    if auth_provider:
        extra_volumes.append(auth_provider.get_path())

    # Mount extra volumes to the container
    volumes: t.List[t.Union[str, dict]] = []

    fs, _ = fsspec.core.url_to_fs(base_path)
    if isinstance(fs, LocalFileSystem):
        logging.info(f"Using local base path: {base_path}")
        logging.info(
            "This directory will be mounted to /artifacts in the container.",
        )
        data_directory_path = Path(base_path).resolve()
        container_path = os.path.join("/", data_directory_path.name)

        command = ["--base_path", f"{container_path}"]

        if extra_volumes:
            volumes.extend(extra_volumes)

        # Mount the local base path to the container
        volume = DockerVolume(
            type="bind",
            source=str(data_directory_path),
            target=f"/{data_directory_path.stem}",
        )
        volumes.append(asdict(volume))

    else:
        if not extra_volumes:
            raise RuntimeError(
                None,
                f"Specified base path `{base_path}`, Please provide valid credentials when using"
                f" a remote base path",
            )

        logging.info(f"Using remote base path: {base_path}")

        # Mount extra volumes to the container
        volumes.extend(extra_volumes)

        # Add the remote base path as a container argument
        command = ["--base_path", base_path]

    services = {
        f"{_get_service_name(container)}": {
            "command": command,
            "volumes": volumes,
            "ports": [f"{port}:8501"],
            "image": f"{shlex.quote(container)}:{shlex.quote(tag)}",
        },
    }

    return {
        "name": "explorer_app",
        "version": "3.8",
        "services": services,
    }


# type: ignore
def run_explorer_app(  # type: ignore  # noqa: PLR0913
    base_path: str,
    port: int = PORT,
    container: str = CONTAINER,
    output_path: str = OUTPUT_PATH,
    tag: t.Optional[str] = None,
    extra_volumes: t.Union[t.Optional[list], t.Optional[str]] = None,
    auth_provider: t.Optional[CloudCredentialsMount] = None,
):  # type: ignore
    """
    Run an Explorer App in a Docker container.

    Args:
      base_path: the base path where the Explorer App will be mounted.
      port: The port number to expose the Explorer App. Default is 8501.
      container: The name of the Docker container. Default is "fndnt/data_explorer".
      output_path: The path to the Docker Compose specification. Default is
       ".fondant/explorer-compose.yaml".
      tag: The tag/version of the Docker container. Default is "latest".
      extra_volumes: Extra volumes to mount in containers. You can use the --extra-volumes flag
      to specify extra volumes to mount in the containers this can be used:
        - to mount data directories to be used by the pipeline (note that if your pipeline's
            base_path is local it will already be mounted for you).
        - to mount cloud credentials
      auth_provider: The cloud provider to use for authentication. Default is None.
    """
    os.makedirs(".fondant", exist_ok=True)

    explorer_app_spec = _generate_explorer_spec(
        base_path=base_path,
        port=port,
        container=container,
        tag=tag,
        extra_volumes=extra_volumes,
        auth_provider=auth_provider,
    )

    with open(output_path, "w") as outfile:
        yaml.dump(explorer_app_spec, outfile, default_flow_style=False)

    cmd = [
        "docker",
        "compose",
        "-f",
        output_path,
        "up",
        "--build",
        "--pull",
        "always",
        "--remove-orphans",
        "--detach",
    ]

    try:
        subprocess.check_call(  # nosec
            cmd,
            stdout=subprocess.PIPE,
            env=dict(os.environ, DOCKER_DEFAULT_PLATFORM="linux/amd64"),
        )
    except subprocess.CalledProcessError as e:
        raise SystemExit(e.returncode)

    logging.info(
        f"Running image from registry '{container}' with tag '{tag}' on port '{port}'",
    )
    logging.info(f"Access the explorer at http://localhost:{port}")


def stop_explorer_app(
    output_path: str = OUTPUT_PATH,
):
    """
    Stop the Explorer App.

    Args:
        output_path: The path to save the Docker Compose specification. Default is
        ".fondant/explorer-compose.yaml".
    """
    cmd = [
        "docker",
        "compose",
        "-f",
        output_path,
        "stop",
    ]

    try:
        subprocess.check_call(cmd, stdout=subprocess.PIPE)  # nosec
    except subprocess.CalledProcessError as e:
        raise SystemExit(e.returncode)

    # check if the container is running
    logging.info(
        "Explorer app stopped successfully",
    )



import json
import typing as t
from abc import abstractmethod
from dataclasses import dataclass

import yaml

from fondant.core.exceptions import InvalidPipelineDefinition


@dataclass
class Accelerator:
    """
    Represents a hardware accelerator configuration.

    Args:
        type: Type of the accelerator.
        number: The number of the accelerator.
    """

    type: str
    number: int


@dataclass
class ComponentConfigs:
    """
    Represents the configurations for a component.

    Args:
        image: The Docker image for the component.
        arguments: Arguments to be passed to the component.
        dependencies: List of dependencies required for the component.
        accelerators: List of hardware accelerators for the component.
        cpu_request: CPU request for the component.
        cpu_limit: CPU limit for the component.
        memory_request: Memory request for the component.
        memory_limit: Memory limit for the component.
    """

    image: t.Optional[str] = None
    command: t.Optional[t.List[str]] = None
    arguments: t.Optional[t.Dict[str, t.Any]] = None
    dependencies: t.Optional[t.List[str]] = None
    accelerators: t.Optional[t.List[Accelerator]] = None
    cpu_request: t.Optional[str] = None
    cpu_limit: t.Optional[str] = None
    memory_request: t.Optional[str] = None
    memory_limit: t.Optional[str] = None


@dataclass
class KubeflowComponentConfig(ComponentConfigs):
    """
    Represents Kubeflow-specific configurations for a component.

    Args:
        node_pool_label: Label for the node pool.
        node_pool_name: Name of the node pool.
    """

    node_pool_label: t.Optional[str] = None
    node_pool_name: t.Optional[str] = None


@dataclass
class DockerComponentConfig(ComponentConfigs):
    """
    Represents Docker-specific configurations for a component.

    Args:
        context: The context for the Docker component.
        volumes: List of volumes for the Docker component.
        ports: List of ports for the Docker component.
    """

    context: t.Optional[str] = None
    volumes: t.Optional[t.List[t.Union[str, dict]]] = None
    ports: t.Optional[t.List[t.Union[str, dict]]] = None


@dataclass
class PipelineConfigs:
    """
    Represents the configurations for a pipeline.

    Args:
        pipeline_name: Name of the pipeline.
        pipeline_description: Description of the pipeline.
    """

    pipeline_name: str
    pipeline_version: str
    pipeline_description: t.Optional[str] = None

    @classmethod
    @abstractmethod
    def from_spec(cls, spec_path: str) -> "PipelineConfigs":
        """Get pipeline configs from a pipeline specification."""


@dataclass
class DockerComposeConfigs(PipelineConfigs):
    """
    Represents Docker-specific configurations for a pipeline.

    Args:
       component_configs: Dictionary of Docker component configurations for the pipeline.
    """

    component_configs: t.Optional[t.Dict[str, DockerComponentConfig]] = None

    @classmethod
    def from_spec(cls, spec_path: str) -> "DockerComposeConfigs":
        """Get pipeline configs from a pipeline specification."""
        with open(spec_path) as file_:
            specification = yaml.safe_load(file_)

        components_configs_dict = {}

        pipeline_description = None
        # Iterate through each service
        for component_name, component_configs in specification["services"].items():
            # Get arguments from command
            command_list = component_configs.get("command", [])
            arguments = {}
            for i in range(0, len(command_list), 2):
                arguments[command_list[i].lstrip("-")] = command_list[i + 1]

            # Get accelerator name and number of accelerators
            resources = component_configs.get("deploy", {}).get("resources", {})
            devices = resources.get("reservations", {}).get("devices", {})

            accelerator_list = []
            if devices:
                for device in devices:
                    accelerator = Accelerator(
                        type=device["capabilities"][0],
                        number=device["count"],
                    )
                    accelerator_list.append(accelerator)

            component_config = DockerComponentConfig(
                image=component_configs.get("image", None),
                arguments=arguments,
                dependencies=list(component_configs.get("depends_on", {}).keys()),
                accelerators=accelerator_list,
                context=component_configs.get("build", {}).get("context", None),
                ports=component_configs.get("ports", None),
                volumes=component_configs.get("volumes", None),
                cpu_request=None,
                cpu_limit=None,
                memory_request=None,
                memory_limit=None,
            )
            components_configs_dict[component_name] = component_config
            pipeline_description = component_configs.get("labels", {}).get(
                "pipeline_description",
                "No description provided",
            )

        return cls(
            pipeline_name=specification["name"],
            pipeline_version=specification["version"],
            pipeline_description=pipeline_description,
            component_configs=components_configs_dict,
        )


@dataclass
class KubeflowPipelineConfigs(PipelineConfigs):
    """
    Represents Kubeflow-specific configurations for a pipeline.

    Args:
        component_configs: Dictionary of Kubeflow component configurations for the pipeline.
    """

    component_configs: t.Optional[t.Dict[str, KubeflowComponentConfig]] = None

    @classmethod
    def from_spec(cls, spec_path: str) -> "KubeflowPipelineConfigs":
        """Get pipeline configs from a pipeline specification."""
        # Two specs are present and loaded in the yaml file (component spec and k8s specs)
        k8_specification = {}
        specification = {}

        with open(spec_path) as file_:
            for spec in yaml.load_all(file_, Loader=yaml.FullLoader):
                if "deploymentSpec" in spec:
                    specification = spec
                elif "platforms" in spec:
                    k8_specification = spec["platforms"]["kubernetes"][
                        "deploymentSpec"
                    ]["executors"]

        if not specification:
            msg = "No component specification found in the pipeline specification"
            raise InvalidPipelineDefinition(msg)
        components_configs_dict = {}

        # Iterate through each service
        for sanitized_component_name, component_configs in specification["root"]["dag"][
            "tasks"
        ].items():
            # Get arguments from command
            arguments = {
                arg_name: arg_value["runtimeValue"]["constant"]
                for arg_name, arg_value in component_configs["inputs"][
                    "parameters"
                ].items()
            }

            component_name = json.loads(arguments["metadata"])["component_id"]
            # Get accelerator name and number of accelerators
            container_spec = specification["deploymentSpec"]["executors"][
                f"exec-{sanitized_component_name}"
            ]["container"]
            resources = component_configs.get("resources", {})
            devices = resources.get("accelerator", {})
            accelerator_list = []

            if devices:
                for device in devices:
                    accelerator = Accelerator(
                        type=device["accelerator"]["type"],
                        number=device["accelerator"]["count"],
                    )
                    accelerator_list.append(accelerator)

            # Get node pool name and label
            node_pool_label = None
            node_pool_name = None
            if k8_specification:
                node_pool_dict = (
                    k8_specification.get(f"exec-{sanitized_component_name}", {})
                    .get("nodeSelector", {})
                    .get("labels", {})
                )
                if node_pool_dict:
                    node_pool_label = list(node_pool_dict.keys())[0]
                    node_pool_name = list(node_pool_dict.values())[0]

            dependencies = component_configs.get("dependentTasks", [])

            if dependencies:
                dependencies = [
                    dependency.replace("-", "_") for dependency in dependencies
                ]

            component_config = KubeflowComponentConfig(
                image=container_spec.get("image"),
                command=container_spec.get("command"),
                arguments=arguments,
                dependencies=dependencies,
                accelerators=accelerator_list,
                cpu_request=component_configs.get("cpuRequest", None),
                cpu_limit=component_configs.get("cpuLimit", None),
                memory_request=component_configs.get("memoryRequest", None),
                memory_limit=component_configs.get("memoryLimit", None),
                node_pool_name=node_pool_name,
                node_pool_label=node_pool_label,
            )
            components_configs_dict[component_name] = component_config

        pipeline_info = specification["pipelineInfo"]

        return cls(
            pipeline_name=pipeline_info["name"],
            pipeline_version=specification["sdkVersion"],
            pipeline_description=pipeline_info.get("description", None),
            component_configs=components_configs_dict,
        )


@dataclass
class VertexPipelineConfigs(KubeflowPipelineConfigs):
    pass



"""Module holding implementation to build Fondant components, used by the `fondant build`
command.
"""

import logging
import re
import typing as t
from pathlib import Path

import docker

from fondant.pipeline import ComponentOp

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def build_component(  # ruff: noqa: PLR0912, PLR0915
    component_dir: Path,
    *,
    tag: t.Optional[str],
    build_args: t.List[str],
    labels: t.List[str],
    nocache: bool = False,
    pull: bool = False,
    target: t.Optional[str] = None,
) -> None:
    component_op = ComponentOp.from_component_yaml(component_dir)
    component_spec = component_op.component_spec

    if component_op.dockerfile_path is None:
        msg = (
            f"Could not detect a `Dockerfile` in {component_dir}. Please make sure it is placed "
            f"at the root of your component_dir and named `Dockerfile`."
        )
        raise SystemExit(msg)

    if tag is None:
        logger.info("No tag provided. Extracting image name from `component_spec.yaml`")
        full_image_name = component_spec.image
    elif ":" in tag:
        logger.info("Detected `:` in tag")
        full_image_name = tag
    else:
        logger.info("Did not detect `:` in tag")
        logger.info("Extracting image name from `component_spec.yaml`")
        repository = component_spec.image.split(":")[0]
        full_image_name = f"{repository}:{tag}"

    logger.info(f"Assuming full image name: {full_image_name}")

    logger.info("Building image...")

    # Convert build args from ["key=value", ...] to {"key": "value", ...}
    build_kwargs = {}
    for arg in build_args:
        k, v = arg.split("=", 1)
        build_kwargs[k] = v

    # Convert label args from ["key=value", ...] to {"key": "value", ...}
    label_kwargs = {}
    for arg in labels:
        k, v = arg.split("=", 1)
        label_kwargs[k] = v

    try:
        docker_client = docker.from_env()
    except docker.errors.DockerException:
        for url in [
            "/var/run/docker.sock",
            Path.home() / ".docker/desktop/docker.sock",
            Path.home() / ".docker/run/docker.sock",
        ]:
            base_url = f"unix://{url}"
            try:
                docker_client = docker.DockerClient(base_url=base_url)
                break
            except docker.errors.DockerException:
                continue
        else:
            msg = "Could not connect to docker daemon, is it running?"
            raise SystemExit(msg)

    logs = docker_client.api.build(
        path=str(component_dir),
        tag=full_image_name,
        buildargs=build_kwargs,
        nocache=nocache,
        pull=pull,
        target=target,
        decode=True,
        labels=label_kwargs,
    )

    for chunk in logs:
        if "stream" in chunk:
            for line in chunk["stream"].splitlines():
                logger.info(line)

    logger.info("Pushing image...")
    repository, tag = full_image_name.split(":")
    logs = docker_client.api.push(repository, tag=tag, stream=True, decode=True)

    for chunk in logs:
        if "error" in chunk:
            logger.error("Push failed:")
            raise SystemExit(chunk["error"])
        message = chunk.get("status", "")
        if "progress" in chunk:
            message += " | " + chunk["progress"]
        logger.info(message)

    logger.info("Updating image name in component_spec")
    # Read and write with `re.sub` to prevent reformatting of file with yaml
    with open(component_dir / component_op.COMPONENT_SPEC_NAME, "r+") as f:
        content = f.read()
        f.seek(0)
        content = re.sub(r"image: [^\n]*", f"image: {full_image_name}", content)
        f.write(content)
        f.truncate()

    logger.info("Done")



"""This module defines classes to represent a Fondant manifest."""
import copy
import json
import pkgutil
import types
import typing as t
from dataclasses import asdict, dataclass
from pathlib import Path

import jsonschema.exceptions
from fsspec import open as fs_open
from jsonschema import Draft4Validator
from referencing import Registry, Resource
from referencing.jsonschema import DRAFT4

from fondant.core.component_spec import OperationSpec
from fondant.core.exceptions import InvalidManifest
from fondant.core.schema import Field, Type


@dataclass
class Metadata:
    """
    Class representing the Metadata of the manifest.

    Args:
        base_path: the base path used to store the artifacts
        pipeline_name: the name of the pipeline
        run_id: the run id of the pipeline
        component_id: the name of the component
        cache_key: the cache key of the component.
    """

    base_path: str
    pipeline_name: str
    run_id: str
    component_id: t.Optional[str]
    cache_key: t.Optional[str]

    def to_dict(self):
        return asdict(self)

    def to_json(self):
        return json.dumps(self.to_dict())

    @classmethod
    def from_dict(cls, data_dict):
        return cls(**data_dict)


class Manifest:
    """
    Class representing a Fondant manifest.

    Args:
        specification: The manifest specification as a Python dict
    """

    def __init__(self, specification: t.Dict[str, t.Any]) -> None:
        self._specification = copy.deepcopy(specification)
        self._validate_spec()

    def _validate_spec(self) -> None:
        """Validate a manifest specification against the manifest schema.

        Raises: InvalidManifest when the manifest is not valid.
        """
        spec_data = pkgutil.get_data("fondant", "core/schemas/manifest.json")

        if spec_data is None:
            msg = "schemas/manifest.json not found"
            raise FileNotFoundError(msg)

        spec_str = spec_data.decode("utf-8")
        spec_schema = json.loads(spec_str)

        base_uri = Path(__file__).parent / "schemas"

        def retrieve_from_filesystem(uri: str) -> Resource:
            path = base_uri / uri
            contents = json.loads(path.read_text())
            return Resource.from_contents(contents, default_specification=DRAFT4)

        registry = Registry(retrieve=retrieve_from_filesystem)  # type: ignore
        validator = Draft4Validator(spec_schema, registry=registry)  # type: ignore

        try:
            validator.validate(self._specification)
        except jsonschema.exceptions.ValidationError as e:
            raise InvalidManifest.create_from(e)

    @classmethod
    def create(
        cls,
        *,
        pipeline_name: str,
        base_path: str,
        run_id: str,
        component_id: t.Optional[str] = None,
        cache_key: t.Optional[str] = None,
    ) -> "Manifest":
        """Create an empty manifest.

        Args:
            pipeline_name: the name of the pipeline
            base_path: The base path of the manifest
            run_id: The id of the current pipeline run
            component_id: The id of the current component being executed
            cache_key: The component cache key
        """
        metadata = Metadata(
            pipeline_name=pipeline_name,
            base_path=base_path,
            run_id=run_id,
            component_id=component_id,
            cache_key=cache_key,
        )

        specification = {
            "metadata": metadata.to_dict(),
            "index": {"location": f"/{run_id}/{component_id}"},
            "fields": {},
        }
        return cls(specification)

    @classmethod
    def from_file(cls, path: t.Union[str, Path]) -> "Manifest":
        """Load the manifest from the file specified by the provided path."""
        with fs_open(path, encoding="utf-8") as file_:
            specification = json.load(file_)
            return cls(specification)

    def to_file(self, path: t.Union[str, Path]) -> None:
        """Dump the manifest to the file specified by the provided path."""
        with fs_open(path, "w", encoding="utf-8", auto_mkdir=True) as file_:
            json.dump(self._specification, file_)

    def copy(self) -> "Manifest":
        """Return a deep copy of itself."""
        return self.__class__(copy.deepcopy(self._specification))

    @property
    def metadata(self) -> t.Dict[str, t.Any]:
        return self._specification["metadata"]

    @property
    def index(self) -> Field:
        return Field(name="id", location=self._specification["index"]["location"])

    def update_metadata(self, key: str, value: t.Any) -> None:
        self.metadata[key] = value

    @property
    def base_path(self) -> str:
        return self.metadata["base_path"]

    def get_field_location(self, field_name: str):
        """Return absolute path to the field location."""
        if field_name == "id":
            return f"{self.base_path}/{self.pipeline_name}{self.index.location}"
        if field_name not in self.fields:
            msg = f"Field {field_name} is not available in the manifest."
            raise ValueError(msg)

        field = self.fields[field_name]
        return f"{self.base_path}/{self.pipeline_name}{field.location}"

    @property
    def run_id(self) -> str:
        return self.metadata["run_id"]

    @property
    def component_id(self) -> str:
        return self.metadata["component_id"]

    @property
    def pipeline_name(self) -> str:
        return self.metadata["pipeline_name"]

    @property
    def cache_key(self) -> str:
        return self.metadata["cache_key"]

    @property
    def fields(self) -> t.Mapping[str, Field]:
        """The fields of the manifest as an immutable mapping."""
        return types.MappingProxyType(
            {
                name: Field(
                    name=name,
                    type=Type.from_dict(field),
                    location=field["location"],
                )
                for name, field in self._specification["fields"].items()
            },
        )

    def add_or_update_field(self, field: Field, overwrite: bool = False):
        """Add or update field to manifest."""
        if field.name == "index":
            self._add_or_update_index(field, overwrite=True)
        elif overwrite is False and field.name in self._specification["fields"]:
            msg = (
                f"A field with name {field.name} already exists. Set overwrite to true, "
                f"if you want to update the field."
            )
            raise ValueError(msg)
        else:
            self._specification["fields"][field.name] = {
                "location": field.location,
                **field.type.to_dict(),
            }

    def _add_or_update_index(self, field: Field, overwrite: bool = True):
        """Add or update the manifest index."""
        if overwrite is False:
            msg = (
                "The index already exists. Set overwrite to true, "
                "if you want to update the index."
            )
            raise ValueError(msg)

        if field.name != "index":
            msg = (
                f"The field name is {field.name}. If you try to update the index, "  # nosec B608
                f"set the field name to `index`."
            )
            raise ValueError(msg)

        self._specification["index"] = {
            "location": field.location,
        }

    def remove_field(self, name: str) -> None:
        if name not in self._specification["fields"]:
            msg = f"Field {name} not found in specification"
            raise ValueError(msg)

        del self._specification["fields"][name]

    def evolve(  # : PLR0912 (too many branches)
        self,
        operation_spec: OperationSpec,
        *,
        run_id: str,
    ) -> "Manifest":
        """Evolve the manifest based on the component spec. The resulting
        manifest is the expected result if the current manifest is provided
        to the component defined by the component spec.

        Args:
            operation_spec: the operation spec
            run_id: the run id to include in the evolved manifest. If no run id is provided,
            the run id from the original manifest is propagated.
        """
        evolved_manifest = self.copy()

        # Update `run_id` and `component_id` in the metadata
        component_id = operation_spec.component_name
        evolved_manifest.update_metadata(key="component_id", value=component_id)
        evolved_manifest.update_metadata(key="run_id", value=run_id)

        # Update index location as this is always rewritten
        evolved_manifest.add_or_update_field(
            Field(name="index", location=f"/{run_id}/{component_id}"),
        )

        # Remove all previous fields if the component changes the index
        if operation_spec.previous_index:
            for field_name in evolved_manifest.fields:
                evolved_manifest.remove_field(field_name)

        # Add or update all produced fields defined in the component spec
        for name, field in operation_spec.produces_to_dataset.items():
            # If field was not part of the input manifest, add field to output manifest.
            # If field was part of the input manifest and got produced by the component, update
            # the manifest field.
            field.location = f"/{run_id}/{component_id}"
            evolved_manifest.add_or_update_field(field, overwrite=True)

        return evolved_manifest

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._specification!r})"



import json
import logging
import os
import re
import shlex
import tempfile
import textwrap
import typing as t
from abc import ABC, abstractmethod
from dataclasses import asdict
from pathlib import Path
from urllib.parse import urlparse

import yaml
from fsspec.registry import known_implementations

from fondant.core.component_spec import ComponentSpec
from fondant.core.exceptions import InvalidPipelineDefinition
from fondant.core.manifest import Metadata
from fondant.core.schema import CloudCredentialsMount, DockerVolume
from fondant.pipeline import (
    VALID_ACCELERATOR_TYPES,
    VALID_VERTEX_ACCELERATOR_TYPES,
    Image,
    Pipeline,
)

logger = logging.getLogger(__name__)

# export DASK_DIAGNOSTICS_PORT="" to get a dynamic port assigned
DASK_DIAGNOSTICS_PORT = os.environ.get("DASK_DIAGNOSTICS_PORT", ":8787")

KubeflowCommandArguments = t.List[t.Union[str, t.Dict[str, str]]]


class Compiler(ABC):
    """Abstract base class for a compiler."""

    @abstractmethod
    def compile(self, *args, **kwargs) -> None:
        """Abstract method to invoke compilation."""

    @abstractmethod
    def _set_configuration(self, *args, **kwargs) -> None:
        """Abstract method to set pipeline configuration."""

    def log_unused_configurations(self, **kwargs):
        """Log configurations that are set but will be unused."""
        for config_name, config_value in kwargs.items():
            if config_value is not None:
                logger.warning(
                    f"Configuration `{config_name}` is set with `{config_value}` but has no effect"
                    f" for runner `{self.__class__.__name__}`.",
                )

    @staticmethod
    def _build_entrypoint(image: Image) -> t.List[str]:
        """Build the entrypoint to execute the provided image."""
        if not image.script:
            # Not a lightweight python component
            return ["fondant", "execute", "main"]

        command = ""
        if image.extra_requires:
            requirements = "\n".join(image.extra_requires)
            command += textwrap.dedent(
                f"""\
                printf {shlex.quote(requirements)} > 'requirements.txt'
                python3 -m pip install -r requirements.txt
            """,
            )

        command += textwrap.dedent(
            f"""\
            printf {shlex.quote(image.script)} > 'main.py'
            fondant execute main "$@"
        """,
        )

        return [
            "sh",
            "-ec",
            command,
            "--",  # All arguments provided after this will be passed to `fondant execute main`
        ]


class DockerCompiler(Compiler):
    """Compiler that creates a docker-compose spec from a pipeline."""

    def compile(
        self,
        pipeline: Pipeline,
        *,
        output_path: str = "docker-compose.yml",
        extra_volumes: t.Union[t.Optional[list], t.Optional[str]] = None,
        build_args: t.Optional[t.List[str]] = None,
        auth_provider: t.Optional[CloudCredentialsMount] = None,
    ) -> None:
        """Compile a pipeline to docker-compose spec and save it to a specified output path.

        Args:
            pipeline: the pipeline to compile
            output_path: the path where to save the docker-compose spec
            extra_volumes: a list of extra volumes (using the Short syntax:
              https://docs.docker.com/compose/compose-file/05-services/#short-syntax-5)
              to mount in the docker-compose spec.
            build_args: List of build arguments to pass to docker
            auth_provider: The cloud provider to use for authentication. Default is None.

        """
        if extra_volumes is None:
            extra_volumes = []

        if isinstance(extra_volumes, str):
            extra_volumes = [extra_volumes]

        if auth_provider:
            extra_volumes.append(auth_provider.get_path())

        logger.info(f"Compiling {pipeline.name} to {output_path}")

        spec = self._generate_spec(
            pipeline,
            extra_volumes=extra_volumes,
            build_args=build_args or [],
        )

        class NoAliasDumper(yaml.SafeDumper):
            def ignore_aliases(self, data):
                return True

        with open(output_path, "w") as outfile:
            yaml.dump(spec, outfile, Dumper=NoAliasDumper, default_flow_style=False)

        logger.info(f"Successfully compiled to {output_path}")

    @staticmethod
    def _patch_path(base_path: str) -> t.Tuple[str, t.Optional[DockerVolume]]:
        """Helper that checks if the base_path is local or remote,
        if local it patches the base_path and prepares a bind mount
        Returns a tuple containing the path and volume.
        """

        def is_remote_path(path: Path) -> bool:
            """Check if the path is remote."""
            scheme = urlparse(str(path)).scheme

            fsspec_schemes = set(known_implementations.keys()) - {"local", "file"}

            return scheme in fsspec_schemes

        def resolve_local_base_path(base_path: Path) -> Path:
            """Resolve local base path and create base directory if it no exist."""
            p_base_path = base_path.resolve()
            try:
                if p_base_path.exists():
                    logger.info(
                        f"Base path found on local system, setting up {base_path} as mount volume",
                    )
                else:
                    p_base_path.mkdir(parents=True, exist_ok=True)
                    logger.info(
                        f"Base path not found on local system, created base path and setting up "
                        f"{base_path} as mount volume",
                    )
            except Exception as e:
                msg = f"Unable to create and mount local base path. {e}"
                raise ValueError(msg)

            return p_base_path

        p_base_path = Path(base_path)
        if is_remote_path(p_base_path):
            logger.info(f"Base path {base_path} is remote")
            return base_path, None

        p_base_path = resolve_local_base_path(p_base_path)
        volume = DockerVolume(
            type="bind",
            source=str(p_base_path),
            target=f"/{p_base_path.stem}",
        )
        path = f"/{p_base_path.stem}"
        return path, volume

    def _generate_spec(
        self,
        pipeline: Pipeline,
        *,
        extra_volumes: t.List[str],
        build_args: t.List[str],
    ) -> dict:
        """Generate a docker-compose spec as a python dictionary,
        loops over the pipeline graph to create services and their dependencies.
        """
        path, volume = self._patch_path(base_path=pipeline.base_path)
        run_id = pipeline.get_run_id()

        services = {}

        pipeline.validate(run_id=run_id)

        component_cache_key = None

        for component_id, component in pipeline._graph.items():
            component_op = component["operation"]

            component_cache_key = component_op.get_component_cache_key(
                previous_component_cache=component_cache_key,
            )

            metadata = Metadata(
                pipeline_name=pipeline.name,
                run_id=run_id,
                base_path=path,
                component_id=component_id,
                cache_key=component_cache_key,
            )

            logger.info(f"Compiling service for {component_id}")

            entrypoint = self._build_entrypoint(component_op.image)

            # add metadata argument to command
            command = ["--metadata", metadata.to_json()]

            # add in and out manifest paths to command
            command.extend(
                [
                    "--output_manifest_path",
                    f"{path}/{metadata.pipeline_name}/{metadata.run_id}/"
                    f"{component_id}/manifest.json",
                ],
            )

            # add arguments if any to command
            for key, value in component_op.arguments.items():
                if isinstance(value, (dict, list)):
                    command.extend([f"--{key}", json.dumps(value)])
                else:
                    command.extend([f"--{key}", f"{value}"])

            # resolve dependencies
            depends_on = {}
            if component["dependencies"]:
                for dependency in component["dependencies"]:
                    depends_on[dependency] = {
                        "condition": "service_completed_successfully",
                    }
                    # there is only an input manifest if the component has dependencies
                    command.extend(
                        [
                            "--input_manifest_path",
                            f"{path}/{metadata.pipeline_name}/{metadata.run_id}/"
                            f"{dependency}/manifest.json",
                        ],
                    )

            volumes: t.List[t.Union[str, dict]] = []
            if volume:
                volumes.append(asdict(volume))
            if extra_volumes:
                volumes.extend(extra_volumes)

            ports = [f"8787{DASK_DIAGNOSTICS_PORT}"]

            services[component_id] = {
                "entrypoint": entrypoint,
                "command": command,
                "depends_on": depends_on,
                "volumes": volumes,
                "ports": ports,
                "labels": {
                    "pipeline_description": pipeline.description,
                },
            }

            self._set_configuration(services, component_op, component_id)

            if component_op.dockerfile_path is not None:
                logger.info(
                    f"Found Dockerfile for {component_id}, adding build step.",
                )
                services[component_id]["build"] = {
                    "context": str(component_op.component_dir.absolute()),
                    "args": build_args,
                }
            else:
                services[component_id]["image"] = component_op.component_spec.image

        return {
            "name": pipeline.name,
            "version": "3.8",
            "services": services,
        }

    def _set_configuration(self, services, fondant_component_operation, component_id):
        resources_dict = fondant_component_operation.resources.to_dict()

        accelerator_name = resources_dict.pop("accelerator_name")
        accelerator_number = resources_dict.pop("accelerator_number")

        # Unused configurations
        self.log_unused_configurations(**resources_dict)

        if accelerator_name is not None:
            if accelerator_name not in VALID_ACCELERATOR_TYPES:
                msg = (
                    f"Configured accelerator `{accelerator_name}`"
                    f" is not a valid accelerator type for Docker Compose compiler."
                    f" Available options: {VALID_VERTEX_ACCELERATOR_TYPES}"
                )
                raise InvalidPipelineDefinition(msg)

            if accelerator_name == "GPU":
                services[component_id]["deploy"] = {
                    "resources": {
                        "reservations": {
                            "devices": [
                                {
                                    "driver": "nvidia",
                                    "count": accelerator_number,
                                    "capabilities": ["gpu"],
                                },
                            ],
                        },
                    },
                }
            elif accelerator_name == "TPU":
                msg = "TPU configuration is not yet implemented for Docker Compose "
                raise NotImplementedError(msg)

        return services


class KubeflowComponentSpec:
    """
    Class representing a Kubeflow component specification.

    Args:
        specification: The kubeflow component specification as a Python dict
    """

    def __init__(self, specification: t.Dict[str, t.Any]) -> None:
        self._specification = specification

    @staticmethod
    def convert_arguments(fondant_component: ComponentSpec):
        args = {}
        for arg in fondant_component.args.values():
            arg_type_dict = {}

            # Enable isOptional attribute in spec if arg is Optional and defaults to None
            if arg.optional and arg.default is None:
                arg_type_dict["isOptional"] = True
            if arg.default is not None:
                arg_type_dict["defaultValue"] = arg.default

            args[arg.name] = {
                "parameterType": arg.kubeflow_type,
                "description": arg.description,
                **arg_type_dict,  # type: ignore
            }

        return args

    @classmethod
    def from_fondant_component_spec(
        cls,
        fondant_component: ComponentSpec,
        command: t.List[str],
        image_uri: str,
    ):
        """Generate a Kubeflow component spec from a Fondant component spec."""
        input_definitions = {
            "parameters": {
                **cls.convert_arguments(fondant_component),
            },
        }

        kfp_safe_name = (
            re.sub(
                "-+",
                "-",
                re.sub("[^-0-9a-z]+", "-", fondant_component.safe_name.lower()),
            )
            .lstrip("-")
            .rstrip("-")
        )
        specification = {
            "components": {
                "comp-"
                + kfp_safe_name: {
                    "executorLabel": "exec-" + kfp_safe_name,
                    "inputDefinitions": input_definitions,
                },
            },
            "deploymentSpec": {
                "executors": {
                    "exec-"
                    + kfp_safe_name: {
                        "container": {
                            "command": command,
                            "image": image_uri,
                        },
                    },
                },
            },
            "pipelineInfo": {"name": kfp_safe_name},
            "root": {
                "dag": {
                    "tasks": {
                        kfp_safe_name: {
                            "cachingOptions": {"enableCache": True},
                            "componentRef": {"name": "comp-" + kfp_safe_name},
                            "inputs": {
                                "parameters": {
                                    param: {"componentInputParameter": param}
                                    for param in input_definitions["parameters"]
                                },
                            },
                            "taskInfo": {"name": kfp_safe_name},
                        },
                    },
                },
                "inputDefinitions": input_definitions,
            },
            "schemaVersion": "2.1.0",
            "sdkVersion": "kfp-2.6.0",
        }
        return cls(specification)

    def to_file(self, path: t.Union[str, Path]) -> None:
        """Dump the component specification to the file specified by the provided path."""
        with open(path, "w", encoding="utf-8") as file_:
            yaml.dump(
                self._specification,
                file_,
                indent=4,
                default_flow_style=False,
                sort_keys=False,
            )

    def to_string(self) -> str:
        """Return the component specification as a string."""
        return json.dumps(self._specification)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._specification!r})"


class KubeFlowCompiler(Compiler):
    """Compiler that creates a Kubeflow pipeline spec from a pipeline."""

    def __init__(self):
        self._resolve_imports()

    def _resolve_imports(self):
        """Resolve imports for the Kubeflow compiler."""
        try:
            import kfp
            import kfp.kubernetes as kfp_kubernetes

            self.kfp = kfp
            self.kfp_kubernetes = kfp_kubernetes

        except ImportError:
            msg = """You need to install kfp to use the Kubeflow compiler,\n
                     you can install it with `pip install fondant[kfp]`"""
            raise ImportError(
                msg,
            )

    def compile(
        self,
        pipeline: Pipeline,
        output_path: str,
    ) -> None:
        """Compile a pipeline to Kubeflow pipeline spec and save it to a specified output path.

        Args:
            pipeline: the pipeline to compile
            output_path: the path where to save the Kubeflow pipeline spec
        """
        run_id = pipeline.get_run_id()
        pipeline.validate(run_id=run_id)
        logger.info(f"Compiling {pipeline.name} to {output_path}")

        def set_component_exec_args(
            *,
            component_op,
            component_args: t.List[str],
            input_manifest_path: bool,
        ):
            """Dump Fondant specification arguments to kfp command executor arguments."""
            dumped_args: KubeflowCommandArguments = []

            component_args.extend(["output_manifest_path", "metadata"])
            if input_manifest_path:
                component_args.append("input_manifest_path")

            for arg in component_args:
                arg_name = arg.strip().replace(" ", "_")
                arg_name_cmd = f"--{arg_name}"

                dumped_args.append(arg_name_cmd)
                dumped_args.append("{{$.inputs.parameters['" + f"{arg_name}" + "']}}")

            component_op.component_spec.implementation.container.args = dumped_args

            return component_op

        @self.kfp.dsl.pipeline(name=pipeline.name, description=pipeline.description)
        def kfp_pipeline():
            previous_component_task = None
            component_cache_key = None

            for component_name, component in pipeline._graph.items():
                logger.info(f"Compiling service for {component_name}")

                component_op = component["operation"]
                # convert ComponentOp to Kubeflow component
                command = self._build_entrypoint(component_op.image)
                image_uri = component_op.image.base_image
                kubeflow_spec = KubeflowComponentSpec.from_fondant_component_spec(
                    component_op.component_spec,
                    command=command,
                    image_uri=image_uri,
                )

                kubeflow_component_op = self.kfp.components.load_component_from_text(
                    text=kubeflow_spec.to_string(),
                )

                # Remove None values from arguments
                component_args = {
                    k: v for k, v in component_op.arguments.items() if v is not None
                }

                component_cache_key = component_op.get_component_cache_key(
                    previous_component_cache=component_cache_key,
                )
                metadata = Metadata(
                    pipeline_name=pipeline.name,
                    run_id=run_id,
                    base_path=pipeline.base_path,
                    component_id=component_name,
                    cache_key=component_cache_key,
                )

                output_manifest_path = (
                    f"{pipeline.base_path}/{pipeline.name}/"
                    f"{run_id}/{component_name}/manifest.json"
                )
                # Set the execution order of the component task to be after the previous
                # component task.
                if component["dependencies"]:
                    for dependency in component["dependencies"]:
                        input_manifest_path = (
                            f"{pipeline.base_path}/{pipeline.name}/"
                            f"{run_id}/{dependency}/manifest.json"
                        )
                        kubeflow_component_op = set_component_exec_args(
                            component_op=kubeflow_component_op,
                            component_args=list(component_args.keys()),
                            input_manifest_path=True,
                        )
                        component_task = kubeflow_component_op(
                            input_manifest_path=input_manifest_path,
                            output_manifest_path=output_manifest_path,
                            metadata=metadata.to_json(),
                            **component_args,
                        )
                        component_task.after(previous_component_task)

                else:
                    kubeflow_component_op = set_component_exec_args(
                        component_op=kubeflow_component_op,
                        component_args=list(component_args.keys()),
                        input_manifest_path=False,
                    )
                    component_task = kubeflow_component_op(
                        metadata=metadata.to_json(),
                        output_manifest_path=output_manifest_path,
                        **component_args,
                    )

                # Set optional configuration
                component_task = self._set_configuration(
                    component_task,
                    component_op,
                )

                # Disable caching
                component_task.set_caching_options(enable_caching=False)

                previous_component_task = component_task

        logger.info(f"Compiling {pipeline.name} to {output_path}")

        self.kfp.compiler.Compiler().compile(kfp_pipeline, output_path)  # type: ignore
        logger.info("Pipeline compiled successfully")

    def _set_configuration(self, task, fondant_component_operation):
        # Used configurations
        resources_dict = fondant_component_operation.resources.to_dict()

        accelerator_number = resources_dict.pop("accelerator_number")
        accelerator_name = resources_dict.pop("accelerator_name")
        node_pool_label = resources_dict.pop("node_pool_label")
        node_pool_name = resources_dict.pop("node_pool_name")
        cpu_request = resources_dict.pop("cpu_request")
        cpu_limit = resources_dict.pop("cpu_limit")
        memory_request = resources_dict.pop("memory_request")
        memory_limit = resources_dict.pop("memory_limit")

        # Unused configurations
        self.log_unused_configurations(**resources_dict)

        # Assign optional specification
        if cpu_request is not None:
            task.set_memory_request(cpu_request)
        if cpu_limit is not None:
            task.set_memory_limit(cpu_limit)
        if memory_request is not None:
            task.set_memory_request(memory_request)
        if memory_limit is not None:
            task.set_memory_limit(memory_limit)
        if accelerator_name is not None:
            if accelerator_name not in VALID_ACCELERATOR_TYPES:
                msg = (
                    f"Configured accelerator `{accelerator_name}` is not a valid accelerator type"
                    f"for Kubeflow compiler. Available options: {VALID_ACCELERATOR_TYPES}"
                )
                raise InvalidPipelineDefinition(msg)

            task.set_accelerator_limit(accelerator_number)
            if accelerator_name == "GPU":
                task.set_accelerator_type("nvidia.com/gpu")
            elif accelerator_name == "TPU":
                task.set_accelerator_type("cloud-tpus.google.com/v3")
        if node_pool_name is not None and node_pool_label is not None:
            task = self.kfp_kubernetes.add_node_selector(
                task,
                node_pool_label,
                node_pool_name,
            )
        return task


class VertexCompiler(KubeFlowCompiler):
    def __init__(self):
        super().__init__()
        self.resolve_imports()

    def resolve_imports(self):
        """Resolve imports for the Vertex compiler."""
        try:
            import kfp

            self.kfp = kfp

        except ImportError:
            msg = """You need to install kfp to use the Vertex compiler,\n
                     you can install it with `pip install fondant[vertex]`"""
            raise ImportError(
                msg,
            )

    def _set_configuration(self, task, fondant_component_operation):
        # Used configurations
        resources_dict = fondant_component_operation.resources.to_dict()

        cpu_limit = resources_dict.pop("cpu_limit")
        memory_limit = resources_dict.pop("memory_limit")
        accelerator_number = resources_dict.pop("accelerator_number")
        accelerator_name = resources_dict.pop("accelerator_name")

        # Unused configurations
        self.log_unused_configurations(**resources_dict)

        # Assign optional specification
        if cpu_limit is not None:
            task.set_cpu_limit(cpu_limit)
        if memory_limit is not None:
            task.set_memory_limit(memory_limit)
        if accelerator_number is not None:
            task.set_accelerator_limit(accelerator_number)
            if accelerator_name not in VALID_VERTEX_ACCELERATOR_TYPES:
                msg = (
                    f"Configured accelerator `{accelerator_name}` is not a valid accelerator type"
                    f"for Vertex compiler. Available options: {VALID_VERTEX_ACCELERATOR_TYPES}"
                )
                raise InvalidPipelineDefinition(msg)

            task.set_accelerator_type(accelerator_name)

        return task


class SagemakerCompiler(Compiler):  # pragma: no cover
    def __init__(self):
        self.ecr_namespace = "fndnt-mirror"
        self._resolve_imports()

    def _resolve_imports(self):
        try:
            import boto3
            import sagemaker
            import sagemaker.processing
            import sagemaker.workflow.pipeline
            import sagemaker.workflow.steps

            self.boto3 = boto3
            self.sagemaker = sagemaker

        except ImportError:
            msg = """You need to install the sagemaker extras to use the sagemaker compiler,\n
                     you can install it with `pip install fondant[sagemaker]`"""
            raise ImportError(
                msg,
            )

    def _build_command(
        self,
        metadata: Metadata,
        arguments: t.Dict[str, t.Any],
        dependencies: t.List[str] = [],
    ) -> t.List[str]:
        # add metadata argument to command
        command = ["--metadata", f"'{metadata.to_json()}'"]

        # add in and out manifest paths to command
        command.extend(
            [
                "--output_manifest_path",
                f"{metadata.base_path}/{metadata.pipeline_name}/{metadata.run_id}/"
                f"{metadata.component_id}/manifest.json",
            ],
        )

        # add arguments if any to command
        for key, value in arguments.items():
            if isinstance(value, (dict, list)):
                command.extend([f"--{key}", f"'{json.dumps(value)}'"])
            else:
                command.extend([f"--{key}", f"'{value}'"])

        # resolve dependencies
        if dependencies:
            for dependency in dependencies:
                # there is only an input manifest if the component has dependencies
                command.extend(
                    [
                        "--input_manifest_path",
                        f"{metadata.base_path}/{metadata.pipeline_name}/{metadata.run_id}/"
                        f"{dependency}/manifest.json",
                    ],
                )

        return command

    def _check_ecr_pull_through_rule(self) -> None:
        logging.info(
            f"Checking existing pull through cache rules for '{self.ecr_namespace}'",
        )

        try:
            self.ecr_client.describe_pull_through_cache_rules(
                ecrRepositoryPrefixes=[self.ecr_namespace],
            )
        except self.ecr_client.exceptions._code_to_exception[
            "PullThroughCacheRuleNotFoundException"
        ]:
            logging.info(
                f"""Pull through cache rule for '{self.ecr_namespace}' not found..
                creating pull through cache rule for '{self.ecr_namespace}'""",
            )

            self.ecr_client.create_pull_through_cache_rule(
                ecrRepositoryPrefix=self.ecr_namespace,
                upstreamRegistryUrl="public.ecr.aws",
            )

            logging.info(
                f"Pull through cache rule for '{self.ecr_namespace}' created successfully",
            )

    def _patch_uri(self, og_uri: str) -> str:
        full_ref, tag = og_uri.split(":")

        ref, *repo = full_ref.split("/")

        def pull_through(repository_name):
            _ = self.ecr_client.batch_get_image(
                repositoryName=repository_name,
                imageIds=[{"imageTag": tag}],
            )
            repo_response = self.ecr_client.describe_repositories(
                repositoryNames=[repository_name],
            )
            return repo_response["repositories"][0]["repositoryUri"] + ":" + tag

        if ref == "fndnt":
            logging.info("Reusable component detected, patching URI")
            uri = pull_through(f"{self.ecr_namespace}/{full_ref}")
        elif ref == "public.ecr.aws":
            logging.info("Public AWS ECR component detected, patching URI")
            uri = pull_through(f"{self.ecr_namespace}/{'/'.join(repo)}")
        else:
            logging.info("Custom component detected")
            # the uri does not need patching
            uri = og_uri
        return uri

    def validate_base_path(self, base_path: str) -> None:
        if not base_path.startswith("s3://"):
            msg = "base_path must be a valid s3 path, starting with s3://"
            raise ValueError(msg)

        if base_path.endswith("/"):
            msg = "base_path must not end with a '/'"
            raise ValueError(msg)

    def compile(
        self,
        pipeline: Pipeline,
        output_path: str,
        *,
        role_arn: t.Optional[str] = None,
    ) -> None:
        """Compile a fondant pipeline to sagemaker pipeline spec and save it
        to a specified output path.

        Args:
            pipeline: the pipeline to compile
            output_path: the path where to save the sagemaker pipeline spec.
            role_arn: the Amazon Resource Name role to use for the processing steps,
            if none provided the `sagemaker.get_execution_role()` role will be used.
        """
        self.ecr_client = self.boto3.client("ecr")
        self.validate_base_path(pipeline.base_path)
        self._check_ecr_pull_through_rule()

        run_id = pipeline.get_run_id()
        path = pipeline.base_path
        pipeline.validate(run_id=run_id)

        component_cache_key = None

        steps: t.List[t.Any] = []

        with tempfile.TemporaryDirectory(dir=os.getcwd()) as tmpdirname:
            for component_name, component in pipeline._graph.items():
                component_op = component["operation"]
                component_cache_key = component_op.get_component_cache_key(
                    previous_component_cache=component_cache_key,
                )

                metadata = Metadata(
                    pipeline_name=pipeline.name,
                    run_id=run_id,
                    base_path=path,
                    component_id=component_name,
                    cache_key=component_cache_key,
                )

                logger.info(f"Compiling service for {component_name}")

                command = self._build_command(
                    metadata,
                    component_op.arguments,
                    component["dependencies"],
                )
                depends_on = [steps[-1]] if component["dependencies"] else []

                image = component_op.image
                entrypoint = self._build_entrypoint(image)

                script_path = self.generate_component_script(
                    entrypoint=entrypoint,
                    command=command,
                    component_name=component_name,
                    directory=tmpdirname,
                )

                if not role_arn:
                    # if no role is provided use the default sagemaker execution role
                    # https://docs.aws.amazon.com/sagemaker/latest/dg/automatic-model-tuning-ex-role.html
                    role_arn = self.sagemaker.get_execution_role()

                resources_dict = self._set_configuration(component_op)

                processor = self.sagemaker.processing.ScriptProcessor(
                    image_uri=self._patch_uri(image.base_image),
                    command=["bash"],
                    instance_count=1,
                    base_job_name=component_name,
                    role=role_arn,
                    **resources_dict,
                )

                step = self.sagemaker.workflow.steps.ProcessingStep(
                    name=component_name,
                    processor=processor,
                    depends_on=depends_on,
                    code=script_path,
                )

                steps.append(step)

            sagemaker_pipeline = self.sagemaker.workflow.pipeline.Pipeline(
                name=pipeline.name,
                steps=steps,
            )
            with open(output_path, "w") as outfile:
                json.dump(
                    json.loads(sagemaker_pipeline.definition()),
                    outfile,
                    indent=4,
                )

    def _set_configuration(
        self,
        fondant_component_operation,
        *args,
        **kwargs,
    ):
        # Used configurations
        resources_dict = fondant_component_operation.resources.to_dict()

        instance_type = resources_dict.pop("instance_type")

        if not instance_type:
            logger.warning(
                """No instance type provided, using default `ml.t3.medium`. See:
            https://docs.aws.amazon.com/sagemaker/latest/dg/notebooks-available-instance-types.html
            for options""",
            )
            instance_type = "ml.t3.medium"

        # Unused configurations
        self.log_unused_configurations(**resources_dict)

        return {"instance_type": instance_type}

    @staticmethod
    def generate_component_script(
        *,
        entrypoint: t.List[str],
        command: t.List[str],
        component_name: str,
        directory: str,
    ) -> str:
        """Generate a bash script for a component to be used as input in a
        sagemaker pipeline step. Returns the path to the script.
        """
        # use shlex.quote to escape special bash chars
        command_string = [arg.replace("'", "") for arg in command]
        cleaned_script = shlex.join([*entrypoint, *command_string])

        with open(f"{directory}/{component_name}.sh", "w") as f:
            f.write(cleaned_script)
        return f"{directory}/{component_name}.sh"



"""This module defines interfaces which components should implement to be executed by fondant."""
import logging
import os
import typing as t
from abc import abstractmethod

import dask
import dask.dataframe as dd
import pandas as pd
from dask.distributed import Client, LocalCluster


class BaseComponent:
    """Base interface for each component, specifying only the constructor.

    Args:
        consume: The schema the component should consume
        produces: The schema the component should produce
        **kwargs: The provided user arguments are passed in as keyword arguments
    """

    def __init__(self):
        self.consumes = None
        self.produces = None

    def setup(self) -> t.Any:
        """Method to do additional component setup. This method can return a state (any object),
        which is passed into the `teardown` method.

        There's two reasons to separate this from `__init__`:
        - It can be overwritten separately
        - The Fondant executor handles the state, which is a good alternative for instance
          attributes if the state is only needed in `setup` / `teardown`, since instance
          attributes need to be pickleable when executing a component method across processes.
        """

    def teardown(self, state: t.Any) -> None:
        """Method called after the component has been executed. The Fondant executor injects the
        state returned by the `setup` method.
        """


class DaskComponent(BaseComponent):
    """Component built on Dask."""

    def __init__(self, **kwargs):
        super().__init__()

    def setup(self) -> t.Any:
        # Don't assume every object is a string
        # https://docs.dask.org/en/stable/changelog.html#v2023-7-1
        dask.config.set({"dataframe.convert-string": False})
        # worker.daemon is set to false because creating a worker process in daemon
        # mode is not possible in our docker container setup.
        dask.config.set({"distributed.worker.daemon": False})
        cluster = LocalCluster(
            processes=True,
            n_workers=os.cpu_count(),
            threads_per_worker=1,
        )
        return Client(cluster)

    def teardown(self, client: t.Any) -> None:
        try:
            client.shutdown()
        except Exception:
            msg = "Caught error while shutting down Client. Exiting anyway."
            logging.exception(msg)
            pass


class DaskLoadComponent(DaskComponent):
    """Component that loads data and returns a Dask DataFrame."""

    @abstractmethod
    def load(self) -> dd.DataFrame:
        pass


class DaskTransformComponent(DaskComponent):
    """Component that transforms an incoming Dask DataFrame."""

    @abstractmethod
    def transform(self, dataframe: dd.DataFrame) -> dd.DataFrame:
        """
        Abstract method for applying data transformations to the input dataframe.

        Args:
            dataframe: A Dask dataframe containing the data specified in the `consumes` section
                of the component specification
        """


class DaskWriteComponent(DaskComponent):
    """Component that accepts a Dask DataFrame and writes its contents."""

    @abstractmethod
    def write(self, dataframe: dd.DataFrame) -> None:
        pass


class PandasTransformComponent(DaskComponent):
    """Component that transforms the incoming dataset partition per partition as a pandas
    DataFrame.
    """

    @abstractmethod
    def transform(self, dataframe: pd.DataFrame) -> pd.DataFrame:
        """
        Abstract method for applying data transformations to the input dataframe.
        Called once for each partition of the data.

        Args:
            dataframe: A Pandas dataframe containing a partition of the data
        """


Component = t.TypeVar("Component", bound=BaseComponent)
"""Component type which can represents any of the subclasses of BaseComponent"""



"""This module defines exceptions thrown by Fondant."""

from jsonschema.exceptions import ValidationError


class FondantException(Exception):
    """All custom Fondant exception should subclass this one."""


class InvalidManifest(ValidationError, FondantException):
    """Thrown when a manifest cannot be validated against the schema."""


class InvalidComponentSpec(ValidationError, FondantException):
    """Thrown when a component spec cannot be validated against the schema."""


class InvalidPipelineDefinition(ValidationError, FondantException):
    """Thrown when a pipeline definition is invalid."""


class InvalidTypeSchema(ValidationError, FondantException):
    """Thrown when a Type schema definition is invalid."""


class UnsupportedTypeAnnotation(FondantException):
    """Thrown when an unsupported type annotation is encountered during type inference."""


class InvalidLightweightComponent(FondantException):
    """Thrown when a component is not a valid lightweight component."""



"""This module defines classes to represent an Fondant component specification."""
import copy
import json
import pkgutil
import pydoc
import types
import typing as t
from dataclasses import dataclass
from pathlib import Path

import jsonschema.exceptions
import pyarrow as pa
import yaml
from jsonschema import Draft4Validator
from referencing import Registry, Resource
from referencing.jsonschema import DRAFT4

from fondant.core.exceptions import InvalidComponentSpec, InvalidPipelineDefinition
from fondant.core.schema import Field, Type


@dataclass
class Argument:
    """
    Component argument.

    Args:
        name: name of the argument
        description: argument description
        type: the python argument type in str format (str, int, ...)
        default: default value of the argument (defaults to None)
        optional: whether an argument is optional or not (defaults to False)
    """

    name: str
    type: t.Type
    description: t.Optional[str] = None
    default: t.Optional[t.Any] = None
    optional: t.Optional[bool] = False

    def __post_init__(self):
        self.default = None if self.default == "None" else self.default
        self.parser = json.loads if self.type in [dict, list] else self.type

    @property
    def kubeflow_type(self) -> str:
        lookup = {
            str: "STRING",
            int: "NUMBER_INTEGER",
            float: "NUMBER_DOUBLE",
            bool: "BOOLEAN",
            dict: "STRUCT",
            list: "LIST",
        }
        return lookup[self.type]

    def to_spec(self):
        spec = {
            "type": self.type.__name__,
            "description": str(self.description),
        }
        if self.optional:
            spec["default"] = self.default if self.default is not None else "None"

        return spec


class ComponentSpec:
    """
    Class representing a Fondant component specification.

    Args:
        name: The name of the component
        image: The docker image uri to use for the component
        description: The description of the component
            consumes: A mapping containing the fields consumed by the operation. The keys are the
            names of the fields to be received by the component, while the values are the
            type of the field.


            produces: A mapping containing the fields produced by the operation. The keys are the
                names of the fields to be produced by the component, while the values are the
                type of the field to be written

            arguments: A dictionary containing the argument name and value for the operation.

    """

    def __init__(
        self,
        name: str,
        image: str,
        *,
        description: t.Optional[str] = None,
        consumes: t.Optional[t.Mapping[str, t.Union[str, pa.DataType, bool]]] = None,
        produces: t.Optional[t.Mapping[str, t.Union[str, pa.DataType, bool]]] = None,
        previous_index: t.Optional[str] = None,
        args: t.Optional[t.Dict[str, t.Any]] = None,
        tags: t.Optional[t.List[str]] = None,
    ):
        spec_dict: t.Dict[str, t.Any] = {
            "name": name,
            "image": image,
        }

        if description:
            spec_dict["description"] = description

        if tags:
            spec_dict["tags"] = tags

        if consumes:
            spec_dict["consumes"] = consumes

        if produces:
            spec_dict["produces"] = produces

        if previous_index:
            spec_dict["previous_index"] = previous_index

        if args:
            spec_dict["args"] = args

        self._specification = spec_dict
        self._validate_spec()

    def _validate_spec(self) -> None:
        """Validate a component specification against the component schema.

        Raises: InvalidComponent when the component specification is not valid.
        """
        spec_data = pkgutil.get_data("fondant", "core/schemas/component_spec.json")

        if spec_data is None:
            msg = "component_spec.json not found in fondant schema"
            raise FileNotFoundError(msg)

        spec_str = spec_data.decode("utf-8")
        spec_schema = json.loads(spec_str)

        base_uri = Path(__file__).parent / "schemas"

        def retrieve_from_filesystem(uri: str) -> Resource:
            path = base_uri / uri
            contents = json.loads(path.read_text())
            return Resource.from_contents(contents, default_specification=DRAFT4)

        registry = Registry(retrieve=retrieve_from_filesystem)  # type: ignore
        validator = Draft4Validator(spec_schema, registry=registry)  # type: ignore

        try:
            validator.validate(self._specification)
        except jsonschema.exceptions.ValidationError as e:
            raise InvalidComponentSpec.create_from(e)

    @classmethod
    def from_file(cls, path: t.Union[str, Path]) -> "ComponentSpec":
        """Load the component spec from the file specified by the provided path."""
        with open(path, encoding="utf-8") as file_:
            specification = yaml.safe_load(file_)
            return cls.from_dict(specification)

    def to_file(self, path) -> None:
        """Dump the component spec to the file specified by the provided path."""
        with open(path, "w", encoding="utf-8") as file_:
            yaml.dump(self._specification, file_)

    @classmethod
    def from_dict(cls, component_spec_dict: t.Dict[str, t.Any]) -> "ComponentSpec":
        """Load the component spec from a dictionary."""
        try:
            return cls(**component_spec_dict)
        except TypeError as e:
            msg = f"Invalid component spec: {e}"
            raise InvalidComponentSpec(msg)

    @property
    def name(self):
        return self._specification["name"]

    @property
    def safe_name(self):
        return self.sanitized_component_name(self._specification["name"])

    def sanitized_component_name(self, name) -> str:
        """Cleans and converts a component name."""
        return name.lower().replace(" ", "_")

    @property
    def description(self):
        return self._specification["description"]

    @property
    def image(self) -> str:
        return self._specification["image"]

    @image.setter
    def image(self, value: str) -> None:
        self._specification["image"] = value

    @property
    def tags(self) -> t.List[str]:
        return self._specification.get("tags", None)

    @property
    def consumes(self) -> t.Mapping[str, Field]:
        """The fields consumed by the component as an immutable mapping."""
        return types.MappingProxyType(
            {
                name: Field(name=name, type=Type.from_dict(field))
                for name, field in self._specification.get("consumes", {}).items()
                if name != "additionalProperties"
            },
        )

    @property
    def consumes_additional_properties(self) -> bool:
        """Returns a boolean indicating whether the component consumes additional properties."""
        return self._specification.get("consumes", {}).get(
            "additionalProperties",
            False,
        )

    @property
    def consumes_is_defined(self) -> bool:
        """Returns a boolean indicating whether the component consumes is defined."""
        return bool(self._specification.get("consumes", {}))

    @property
    def produces_additional_properties(self) -> bool:
        """Returns a boolean indicating whether the component produces additional properties."""
        return self._specification.get("produces", {}).get(
            "additionalProperties",
            False,
        )

    @property
    def produces(self) -> t.Mapping[str, Field]:
        """The fields produced by the component as an immutable mapping."""
        return types.MappingProxyType(
            {
                name: Field(name=name, type=Type.from_dict(field))
                for name, field in self._specification.get("produces", {}).items()
                if name != "additionalProperties"
            },
        )

    def is_generic(self, mapping: str) -> bool:
        """Returns a boolean indicating whether the provided mapping is generic.

        Args:
            mapping: "consumes" or "produces"
        """
        additional_fields = self._specification.get(mapping, {}).get(
            "additionalProperties",
        )

        return bool(additional_fields)

    @property
    def previous_index(self) -> t.Optional[str]:
        return self._specification.get("previous_index")

    @property
    def args(self) -> t.Mapping[str, Argument]:
        args = self.default_arguments
        args.update(
            {
                name: Argument(
                    name=name,
                    description=arg_info.get("description"),
                    type=pydoc.locate(arg_info["type"]),  # type: ignore
                    default=arg_info.get("default"),
                    optional=arg_info.get("default") == "None",
                )
                for name, arg_info in self._specification.get("args", {}).items()
            },
        )
        return types.MappingProxyType(args)

    @property
    def specification(self) -> t.Dict[str, t.Any]:
        return copy.deepcopy(self._specification)

    @property
    def default_arguments(self) -> t.Dict[str, Argument]:
        """Add the default arguments of a fondant component."""
        return {
            "input_manifest_path": Argument(
                name="input_manifest_path",
                description="Path to the input manifest",
                type=str,
                optional=True,
            ),
            "operation_spec": Argument(
                name="operation_spec",
                description="The operation specification as a dictionary",
                type=str,
            ),
            "input_partition_rows": Argument(
                name="input_partition_rows",
                description="The number of rows to load per partition. \
                        Set to override the automatic partitioning",
                type=int,
                optional=True,
            ),
            "cache": Argument(
                name="cache",
                description="Set to False to disable caching, True by default.",
                type=bool,
                default=True,
            ),
            "metadata": Argument(
                name="metadata",
                description="Metadata arguments containing the run id and base path",
                type=str,
            ),
            "output_manifest_path": Argument(
                name="output_manifest_path",
                description="Path to the output manifest",
                type=str,
            ),
        }

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._specification!r})"

    def __eq__(self, other):
        if not isinstance(other, ComponentSpec):
            return False
        return self._specification == other._specification


class OperationSpec:
    """A spec for the operation, which contains the `consumes` and `produces` sections of the
    component spec, updated with the `consumes` and `produces` mappings provided as arguments to
    the operation.
    """

    def __init__(
        self,
        component_spec: ComponentSpec,
        *,
        consumes: t.Optional[t.Dict[str, t.Union[str, pa.DataType]]] = None,
        produces: t.Optional[t.Dict[str, t.Union[str, pa.DataType]]] = None,
    ) -> None:
        self._component_spec = component_spec

        self._mappings = {
            "consumes": consumes,
            "produces": produces,
        }
        self._validate_mappings()

        self._operation_consumes: t.Optional[t.Mapping[str, Field]] = None
        self._consumes_from_dataset: t.Optional[t.Mapping[str, Field]] = None
        self._operation_produces: t.Optional[t.Mapping[str, Field]] = None
        self._produces_to_dataset: t.Optional[t.Mapping[str, Field]] = None

    def to_dict(self) -> dict:
        def _dump_mapping(
            mapping: t.Optional[t.Dict[str, t.Union[str, pa.DataType]]],
        ) -> dict:
            if mapping is None:
                return {}

            serialized_mapping: t.Dict[str, t.Any] = mapping.copy()
            for key, value in mapping.items():
                if isinstance(value, pa.DataType):
                    serialized_mapping[key] = Type(value).to_dict()
            return serialized_mapping

        return {
            "specification": self._component_spec.specification,
            "consumes": _dump_mapping(self._mappings["consumes"]),
            "produces": _dump_mapping(self._mappings["produces"]),
        }

    def to_json(self) -> str:
        specification_dict = self.to_dict()
        return json.dumps(specification_dict)

    @classmethod
    def from_dict(cls, operation_spec_dict: t.Dict[str, t.Any]) -> "OperationSpec":
        def _parse_mapping(
            json_mapping: dict,
        ) -> t.Optional[t.Dict[str, t.Union[str, pa.DataType]]]:
            """Parse a json mapping to a Python mapping with Fondant types."""
            for key, value in json_mapping.items():
                if isinstance(value, dict):
                    json_mapping[key] = Type.from_dict(value).value
            return json_mapping

        return cls(
            component_spec=ComponentSpec.from_dict(
                operation_spec_dict["specification"],
            ),
            consumes=_parse_mapping(operation_spec_dict["consumes"]),
            produces=_parse_mapping(operation_spec_dict["produces"]),
        )

    @classmethod
    def from_json(cls, operation_spec_json: str) -> "OperationSpec":
        return cls.from_dict(json.loads(operation_spec_json))

    def _validate_mappings(self) -> None:
        """Validate received consumes and produces mappings on their types."""
        for name, mapping in self._mappings.items():
            if not mapping:
                continue
            for key, value in mapping.items():
                if not isinstance(value, (str, pa.DataType)):
                    msg = f"Unexpected type {type(value)} received for key {key} in {name} mapping"
                    raise InvalidPipelineDefinition(msg)

    def _dataset_schema_to_operation_schema(self, name: str) -> t.Mapping[str, Field]:
        """Calculate the operations schema based on dataset schema.
        Maps dataset fields to the fields that the operation will receive.
        This is the mapping that the component
        `transform` (or equivalent) method will receive. This is calculated by starting from the
        component spec section, and updating it with any string to type mappings from the
        argument mapping.

        Args:
            name: "consumes" or "produces"
        """
        spec_mapping = getattr(self._component_spec, name)
        args_mapping = self._mappings[name]

        if not args_mapping:
            return spec_mapping

        mapping = dict(spec_mapping)

        for key, value in args_mapping.items():
            if not isinstance(value, pa.DataType):
                continue

            if key not in spec_mapping:
                mapping[key] = Field(name=key, type=Type(value))
            else:
                spec_type = spec_mapping[key].type.value
                if spec_type == value:
                    # Same info in mapping and component spec, let it pass
                    pass
                else:
                    msg = (
                        f"Received pyarrow DataType value {value} for key {key} in the "
                        f"`{name}` argument passed to the operation, but `{key}` is "
                        f"already defined in the `{name}` section of the component spec "
                        f"with type {spec_type}"
                    )
                    raise InvalidPipelineDefinition(msg)

        return types.MappingProxyType(mapping)

    def _operation_schema_to_dataset_schema(self, name: str) -> t.Mapping[str, Field]:
        """
        Maps the operations fields to the dataset fields which will be written in DataIO.
        This is calculated by starting from the "inner mapping" updating it
        with any string to string mappings from the argument mapping.

        Args:
            name: "consumes" or "produces"
        """
        operations_schema = getattr(self, f"operation_{name}")

        # Dataset arguments mapping
        operation_to_dataset_mapping = self._mappings[name]

        if not operation_to_dataset_mapping:
            return operations_schema

        mapping = dict(operations_schema)

        for (
            operations_column_name,
            dataset_column_name_or_type,
        ) in operation_to_dataset_mapping.items():
            # If the value is not a string, it means it's a type, so we skip it
            if not isinstance(dataset_column_name_or_type, str):
                continue

            if operations_column_name in operations_schema:
                mapping[dataset_column_name_or_type] = Field(
                    name=dataset_column_name_or_type,
                    type=mapping.pop(operations_column_name).type,
                )
            else:
                msg = (
                    f"Received a string value for key `{operations_column_name}` in the `{name}` "
                    f"argument passed to the operation, but `{operations_column_name}` is not "
                    f"defined in the `{name}` section of the component spec."
                )
                raise InvalidPipelineDefinition(msg)

        return types.MappingProxyType(mapping)

    @property
    def operation_consumes(self) -> t.Mapping[str, Field]:
        """The operations `consumes` mapping which the component `transform` (or equivalent) method
        will receive.
        """
        if self._operation_consumes is None:
            self._operation_consumes = self._dataset_schema_to_operation_schema(
                "consumes",
            )

        return self._operation_consumes

    @property
    def consumes_from_dataset(self) -> t.Mapping[str, Field]:
        """Defines which fields of the dataset are consumed by the operation."""
        if self._consumes_from_dataset is None:
            self._consumes_from_dataset = self._operation_schema_to_dataset_schema(
                "consumes",
            )

        return self._consumes_from_dataset

    @property
    def operation_produces(self) -> t.Mapping[str, Field]:
        """The operations `produces` mapping which the component `transform` (or equivalent) method
        will receive.
        """
        if self._operation_produces is None:
            self._operation_produces = self._dataset_schema_to_operation_schema(
                "produces",
            )

        return self._operation_produces

    @property
    def produces_to_dataset(self) -> t.Mapping[str, Field]:
        """The produces schema used by data_io to write the dataset."""
        if self._produces_to_dataset is None:
            self._produces_to_dataset = self._operation_schema_to_dataset_schema(
                "produces",
            )

        return self._produces_to_dataset

    @property
    def operation_consumes_to_dataset_consumes(self):
        """
        The consumes name mapping. The key is the name of the field in the operation, value is the
        name of the field in the dataset.

        E.g.:
        {
            "OperationField": "DatasetField"
        }
        """
        return self._mappings["consumes"]

    @property
    def component_name(self) -> str:
        """Get the component name."""
        return self._component_spec.safe_name

    @property
    def previous_index(self) -> t.Optional[str]:
        """The name of the index column of the previous component."""
        return self._component_spec.previous_index

    @property
    def args(self) -> t.Mapping[str, Argument]:
        """The component arguments."""
        return self._component_spec.args

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, OperationSpec):
            return False

        # Compare component_spec attribute
        if self._component_spec != other._component_spec:
            return False

        # Compare mappings attribute
        if self._mappings != other._mappings:
            return False

        return True
